"""Streaming module - SSE helpers and message format utilities."""
from .sse import format_sse_event, format_sse_error
from .message_mappers import map_to_responses_api, clean_output

__all__ = ["format_sse_event", "format_sse_error", "map_to_responses_api", "clean_output"]



import json, asyncio, uuid, tempfile, time
from typing import List
from concurrent.futures import ThreadPoolExecutor

import boto3

try:
    from sagemaker.transformer import Transformer
except ImportError:
    Transformer = None

from animuz_core.utils import init_logger

LOGGER = init_logger(__name__)
LOGGER.setLevel("INFO")

class S3EmbeddingClient():
    def __init__(self) -> None:
        self.runtime = boto3.client("sagemaker-runtime")
        self.model_name = "huggingface-pytorch-inference-2024-05-28-06-25-32-617"
        self.endpoint_name = "bge-m3-serverless"
        self.s3_client = boto3.client("s3")
        self.s3_bucket_name = "animuz-datasource"
        self.s3_prefix = "batch_transform_datasource"
        self.executor = ThreadPoolExecutor(max_workers=1)

    def _get_embedding(self, text: str) -> dict:
        """
        Sends a POST request to the embedding server to get the embedding for the given text.
        
        Args:
            text (str): The text to get the embedding for.
        
        Returns:
            dict: The embedding of the given text in the form of a dictionary.
        """
        payload = json.dumps({"doc_id":1, "document_text": text}) + "\n"
        payload = payload.encode()
        start = time.time()
        response = self.runtime.invoke_endpoint(
            EndpointName=self.endpoint_name,
            ContentType="application/jsonlines",
            Accept="application/jsonlines",
            Body=payload
        )
        LOGGER.debug(json.dumps({
            "timestamp": time.strftime('[%Y-%b-%d %H:%M:%S %z]'),
            "message": "Sagemaker serverless embedding latency",
            "latency": str(time.time() - start),
            "exception": False
        }))
        data = response["Body"].read()
        data = json.loads(data.decode())
        data["embedding"] = data["dense_embeddings"] # keep forgetting to rename the keysss
        data["sparse_embedding"] = data["sparse_embeddings"]
        del data["dense_embeddings"], data["sparse_embeddings"]
        return data
    
    async def get_embedding(self, text: str) -> dict:
        try:
            loop = asyncio.get_event_loop()
            result = await loop.run_in_executor(self.executor, self._get_embedding, text)
        except Exception as e:
            LOGGER.critical(json.dumps({
                "timestamp": time.strftime('[%Y-%b-%d %H:%M:%S %z]'),
                "message": "ERROR Sagemaker serverless embedding",
                "latency": None,
                "exception": True
            }), exc_info=True)
            raise(e)
        return result

    def _batch_get_embedding(self, texts) -> List[dict]:
        start = time.time()
        #upload texts to s3
        object_prefix = self.s3_prefix + "/" + str(uuid.uuid4().hex) + "/" 

        with tempfile.NamedTemporaryFile() as tmp:
            data_jsonl = ""
            for i, text in enumerate(texts):
                record = {"doc_id": i + 1, "document_text": text}
                data_jsonl += json.dumps(record, ensure_ascii=False) + "\n"
            tmp.write(data_jsonl.strip().encode())
            tmp.seek(0)

            with open(tmp.name, 'rb') as f:
                self.s3_client.upload_fileobj(f, self.s3_bucket_name, object_prefix + "input.jsonl")

        #start transform job
        transformer = Transformer(model_name=self.model_name,
            instance_count=1,
            instance_type="ml.m5.2xlarge",
            strategy="MultiRecord",
            accept='application/jsonlines',  # Adjust based on your output format
            assemble_with='Line',  # Common for CSV input,
            output_path = f"s3://{self.s3_bucket_name}/{object_prefix}" 
            )
        
        transformer.transform(
            data=f"s3://{self.s3_bucket_name}/{object_prefix + 'input.jsonl'}",
            content_type='application/jsonlines',  # Specify the content type
            split_type='Line'  # Split type for JSONL input
        )

        transformer.wait()
        LOGGER.debug(json.dumps({
            "timestamp": time.strftime('[%Y-%b-%d %H:%M:%S %z]'),
            "message": "Sagemaker serverless batch embedding latency",
            "latency": str(time.time() - start),
            "exception": False
        }))
        #download result
        output = []
        with tempfile.NamedTemporaryFile() as tmp:
            self.s3_client.download_file(self.s3_bucket_name, f'{object_prefix}' + "input.jsonl.out", tmp.name)

            # Read and process the contents of the temporary file
            with open(tmp.name, 'r') as f:
                contents = f.read()

            for line in contents.splitlines():
                record = json.loads(line)
                record["embedding"] = record["dense_embeddings"]
                record["sparse_embedding"] = record["sparse_embeddings"]
                del record["dense_embeddings"], record["sparse_embeddings"]
                output.append(record)
        return output

    async def batch_get_embedding(self, texts):
        try:
            loop = asyncio.get_event_loop()
            result = await loop.run_in_executor(self.executor, self._batch_get_embedding, texts)
        except Exception as e:
            LOGGER.critical(json.dumps({
                "timestamp": time.strftime('[%Y-%b-%d %H:%M:%S %z]'),
                "message": "ERROR Sagemaker serverless batch embedding",
                "latency": None,
                "exception": True
            }), exc_info=True)
            raise(e)
        return result
    
if __name__ == "__main__":
    async def main():
        client = S3EmbeddingClient()
        text = ['<figure>  \n![](figures/0)  \n<!-- FigureContent="krungsri Capital Securities" -->  \n</figure>  \nA member of MUFG, a global financial group',
 '# Thai Union Group TU  \nEarnings Preview 1 April 2024',
 '## AGRO & FOODS  \nSummary  \nเรามีมุมมองเป็นกลาง คาด TU มีกำไรปกติใน 1Q24F ที่ 1,090 ลบ +35%y-y -7%q-q ภาพ y-y ชัดเจนว่าธุรกิจ petfood เด่นสุดด้วยยอดขายโต 15% GPM ฟื้นตัวจากสินค้าพรีเมียม เพิ่มขึ้นและต้นทุนทูน่าลดลง ธุรกิจอาหารทะเลกระป๋องโตน้อย ราคาทูน่าที่ลดลงทำให้ ราคาขายลด แต่ปริมาณเพิ่มขึ้นชดเชย GPM ดีขึ้นเล็กน้อย ต้นทุนทูน่าต่ำยังสะท้อนไม่ เต็มที่ในไตรมาสนี้ ธุรกิจอาหารทะเลแช่แข็งยอดขายอ่อนแต่ GPM ดีขึ้น Red Lobster ไม่ มีการบันทึกส่วนแบ่งแล้วใน 1Q24F ส่วน 2Q24F คาดกำไรปกติเพิ่ม y-y q-q กลุ่ม Petfood ยังโต กลุ่มอาหารทะเลกระป๋องดีขึ้นอีก กลุ่มอาหารทะเลแช่แข็ง ปริมาณและ GPM ดีขึ้น เพราะปรับโครงสร้างธุรกิจครบปีแล้ว เราแนะนำ Buy TU ด้วย TP 17.5 บาท (P/E 14 เท่า) บนแนวโน้มกำไรปกติปี 24F ฟื้นตัว +23% เพราะธุรกิจทั้งทูน่า/อาหารสัตว์เลี้ยง/อาหาร แช่แข็งจะมีความต้องการฟื้นตัวกลับมาในระดับปกติ หลังจากลูกค้าชะลอการซื้อในปี 23 เพื่อใช้สต๊อกเก่า แนวโน้มต้นทุนทูน่าลดลง',
 '# คาดกำไรปกติ 1Q24F ที่ 1,090 ลบ. +35%y-y -7%q-q  \nคาดกำไรสุทธิ 1,040 ลบ ใน 1Q24F เพิ่ม +2%y-y ฟื้นตัวจากขาดทุนใน 4Q23  \nเราคาดวันที่ 8 พ.ค. TU จะรายงานกำไรสุทธิ 1,040 ลบ.ใน 1Q24F เพิ่ม +2%y-y ฟื้นตัวจากขาดทุน -17,189 ลบ.ใน 4Q23 ส่วนกำไรปกติ 1,090 ลบ. ใน 1Q24F เพิ่ม +35%y-y ลด -7%q-q การเพิ่ม y-y เพราะธุรกิจ petfood มียอดขาย และ GPM เพิ่มจากตลาดฟื้นตัว ลูกค้าต้องการเพิ่มสต๊อก ขณะที่ธุรกิจ ทูน่ากระป๋องมีความต้องการเพิ่มแต่ราคาลด และ GPM ดีขึ้น ธุรกิจอาหารทะเลแช่แข็งทำการลดสินค้า ที่ไม่กำไรออก จึงส่งผลให้ยอดขายลดแต่ GPM ดีขึ้น ส่วนการลด q-q ตามฤดูกาลและธุรกิจทูน่า กระป๋องแบบ OEM ลดลงตามราคาทูน่าปรับลดราคาขายและมาร์จิ้นลง  \nRed Lobster ใน 1Q24F ไม่มีการรับรู้ส่วนแบ่งกำไร/ขาดทุนแล้ว ต่างจาก 1Q23 รับรู้ส่วนแบ่งกำไร 20 ลบ. และ 4Q23 รับรู้ขาดทุน -554 ลบ. หลัง TU ประกาศถอนการลงทุนใน RL  \nรายการพิเศษใน 1Q24F คาดขาดทุน FX -50 ลบ. เทียบกับ 1Q23 กำไร FX 215 ลบ. และ 4Q23 กำไร FX 67 ลบ  \nKey numbers : ยอดขายและ GPM เพิ่ม y-y แต่ลด q-q และไม่บันทึก Red lobster  \nยอดขายคาดที่ 33,898 ลบ. ใน 1Q24F เพิ่ม +4%y-y ลด -4%q-q ยอดขายเพิ่ม y-y โดย 1)ยอดขาย อาหารสัตว์เลี้ยงเพิ่ม +15%y-y ลูกค้ากลับมาซื้อและมีลูกค้า retailer ในสหรัฐเพิ่ม 2) ยอดขายอาหาร ทะเลกระป๋องลด y-y ด้วยราคาขายลดตามราคาทูน่าที่ลด -27%y-y ปริมาณขายเพิ่มขึ้นเพราะราคา จูงใจ 3)ยอดขายอาหารทะเลแช่แข็งลดลง -20%y-y เพราะอยู่ในช่วงปรับโครงสร้างตัดสินค้าที่ไม่ กำไรออกไป 4)ยอดขาย packaging และอาหารพร้อมรับประทานเพิ่มขึ้น ทั้งนี้ราคาทูน่าใน 1Q24 อยู่ที่US$1333/ตัน -27%y-y -13%q-q ยังช่วย GPM ธุรกิจทูน่ากระป๋องได้ไม่มาก แต่คาดว่าจะดีขึ้น ใน 2Q24F  \nGPM คาด 16.8% ใน 1Q24F vs 15.0%/17.8% ใน 1Q23/4Q23 การเพิ่ม y-y โดยธุรกิจอาหารแช่ แข็งปรับโครงสร้างธุรกิจลดสินค้าที่ไม่ทำกำไร ธุรกิจ petfood ปรับเพิ่มราคาขายและสินค้า premium เพิ่มขึ้น ธุรกิจอาหารกระป๋องมีต้นทุนทูน่าลดลง อย่างไรก็ตาม GPM รวมลด q-q ตามฤดูกาลที่ ธุรกิจ petfood มีการขายพรีเมี่ยมลดลง ธุรกิจอาหารทะเลแช่แข็งมีราคาขายลดตามกุ้งที่ราคาลดลงและพ รีเมี่ยมลดลง อย่างไรก็ตามเราคาดธุรกิจทูน่ากระป๋องมี GPM เพิ่ม เพราะปริมาณขายเพิ่มจึงมีต้นทุน คงที่ต่อหน่วยเพิ่มขึ้น  \nRed Lobster (RL) ไม่บันทึกส่วนแบ่งกำไร/ขาดทุนแล้ว ต่างจากบันทึกส่วนแบ่งกำไร(ขาดทุน) +20/ -554 ลบ.ใน 1Q23/4Q23  \n|||\n| - | - |\n| Rating | Buy |\n| Remains ||\n| TP 24F (14xP/E) | THB 17.50 |\n| Remains ||\n| Closing price | THB 14.50 |\n| Upside/Downside | +20.7% |  \nFundamental Investment Analyst on Securities  \nPloenjai Jirajarus Analyst Registration No. 17839 02 638 5000, 02 081 2000 Ploenjai.Jirajarus@krungsricapital.com  \n| Forecasts & Valuation |||||\n| 31 Dec (THB) | 23 | 24F | 25F | 26F |\n| - | - | - | - | - |\n| Revenue (mn) |||| 136,153 141,157 147,216 153,686    |\n| EBITDA (mn) |||| 10,698 11,781 12,024 12,429    |\n| Norm. Profit(mn) || 4,762 5,875  | 6,115 | 6,373 |\n| Net Profit (mn) || 13,933) 5,875  | 6,115 | 6,373 |\n| EPS (THB) | (3.15) | 1.26 | 1.31 | 1.37 |\n| EPS Growth (%) | \'314.29) | Up | 4.09 | 4.21 |\n| PER (x) | (4.60) | 11.49 | 11.04 | 10.59 |\n| BV/share | 12.59 | 13.34 | 14.03 | 14.74 |\n| P/BV (x) | 1.15 | 1.09 | 1.03 | 0.98 |\n| EV/EBITDA (x) :selected: | 11.13 | 9.22 | 8.98 | 8.63 |\n| DPS (THB) | 0.54 | 0.63 | 0.66 | 0.68 |\n| Dividend yield (%) | 3.72 | 4.35 | 4.53 | 4.72 |\n| ROE (%) | (20.01) | 9.73 | 9.60 | 9.52 |\n| Int. bearing D/E (x) :selected: | 1.03 | 0.98 | 0.94 | 0.89 |  \nSource: TU, KCS  \n| THBmn ||||| 1Q23 4Q23 1Q24F %y-y %q-q     |\n| - | - | - | - | - | - |\n| Sales ||||| 32,652 35,529 33,898 3.8% -4.6%     |\n| Gross profit | 4,917 |||| 6,327 5,695 15.8% -10.0%    |\n| % GPM ||| 15.06% 17.81% 16.80%   || 1.7% -1.0%  |\n| EBITDA | 1,725 | 3,124 ||| 2,597 50.5% -16.9%   |\n| % EBITDA ||||| 5.28% 8.79% 7.66% 2.4% -1.1%     |\n| Red Lobster | 20 | (554) || 0 -100%  | Up |\n| Norm profit | 806 | 1,177 ||| 1,090 35.3% -7.4%   |\n| % Norm | 2.47% | 3.31% ||| 3.22% 0.7% -0.1%   |\n| Net profit || 1,021 -17,189  || 1,040 1.8%  | Up |\n| FX gain | 215 | 67 | -50 | | |\n| Extra items || 1 -18,433  | O | | |  \nSource: TU, KCS  \n<!-- PageFooter="IRIS Division" -->  \n<!-- PageFooter="www.krungsricapital.com" -->  \n<!-- PageHeader="Krungsri Capital Securities | Thai Union Group" -->  \n<!-- PageHeader="1 April 2024" -->  \nส่วนแบ่งจากบริษัทย่อยคือ Avanti Feed และ Avanti Frozen กำไร +174 ลบ. ใน 1Q24F ลด -17%y-y เพิ่ม +13%q-q จาก +211/+154 ลบ.ใน 1Q23/4Q23 เพราะการส่งออกกุ้งและการเลี้ยงกุ้งในอินเดียลด y-y จากการแข่งราคาของเอกวาดอร์ แต่ภาพ q-q ดีขึ้นตามฤดูกาลมีการลงกุ้งในฤ ฤดูกาลใหม่',
 '## คาด 2Q24F กำไรปกติดีขึ้น y-y q-q รับรู้ต้นทุนทูน่าลดลง  \nคาดกำไรปกติใน 2Q24F ราว 1,300-1,400 ลบ เพิ่มขึ้น y-y q-q vs 1,275 /1,090 ลบ ใน 2Q23/1Q24F ภาพ y-y q-q 1) ธุรกิจ petfood ยอดขายและ GPM ดีขึ้น ปริมาณขายฟื้นตัว สินค้าพรีเมียมเพิ่ม และต้นทุนทูน่าลดลง 2) ธุรกิจอาหารทะเลกระป๋อง ปริมาณขายเพิ่มขึ้น GPM เพิ่มขึ้นจากต้นทุนทู น่าลดลง 3) ธุรกิจอาหารทะเลแช่แข็ง ปริมาณขายและ GPM ดีขึ้น โครงการลดสินค้าไม่กำไรทำให้ GPM ดีขึ้น และยอดขายฟื้นตัว',
 '# ราคาทูน่าเดือนมีนาคมอยู่ที่ US$1300/ton (-33%y-y คงที่m-m) คาดราคาทรงตัวระดับ US$1400-1700 ใน 2Q-4Q24  \nราคาทูน่าเดือนมีนาคม 2024 ที่ US$1300/ton ลดลง -33%y-y คงที่ m-m จาก US$1980 เดือนมีค.23 และ US$1300 เดือนกพ.24 ทำให้ราคาเฉลี่ย 1Q24F อยู่ที่ US$1333 ลด -27%y-y -13%q-q จาก US$1820 ใน 1Q23 และ US$1533 ใน 4Q23 ผู้บริหาร TU ให้ความเห็นว่าราคาทูน่าในปี 2024 จะ อยู่ระหว่าง US$1400-1700 ลดลงจากปี 2023 ที่เฉลี่ย US$1840 อย่างไรก็ตามเราคาดว่าราคาทูน่าจะทรงตัวถึงเพิ่มขึ้นเล็กน้อยในเดือนเมษายน เป็น ต้นไป เพราะหากราคาทูน่าต่ำมาก ชาวประมงก็ไม่อยากออกจับปลาที่ให้ผลตอบแทนน้อย',
 '## แนะนำ Buy ราคาเป้าหมาย 17.50 บาท (พีอี 14 เท่า= mean)  \nเราคงประมาณการกำไรปกติปี 2024/25F ไว้ที่เดิม เนื่องจาก preview GPM และ ยอดขายใน 1Q24F ถึงแม้จะต่ำกว่าคาดเดิมบ้าง แต่จะดีขึ้นใน 2Q24F โดยเฉพาะในส่วนของยอดขายและ GPM ของอาหารทะเลกระป๋อง  \nเราคงแนะนำ Buy ราคาเป้าหมาย 17.50 บาท เพราะแนวโน้มปี 2024F ผลการดำเนินงาน TU จะมีกำไรปกติเติบโต +23% เกิดจากธุรกิจทั้งทูน่ากระป๋อง Petfood จะมีความต้องการฟื้นตัวกลับมาในระดับปกติ หลังจากลูกค้าชะลอการซื้อในปี 2023 เพื่อใช้สต๊อกเก่า ประกอบกับคาดราคาทูน่าจะลดลงในปี 2024F ส่วนธุรกิจอาหารทะเลแช่แข็งคาดฟื้นตัวหลังปรับโครงสร้างธุรกิจ ยกเลิกสินค้ากลุ่มที่ไม่มีกำไรออกไป ทำให้ GPM ของธุรกิจอาหารแช่แข็งดีขึ้น นอกจากนี้การถอนการลงทุนใน Red Lobster ทำให้ TU สามารถfocus ในการขยายธุรกิจหลักของบริษัทได้ เราให้ TU เป็น Top pick ของกลุ่มเกษตรคู่ กับ GFPT เพราะการฟื้นตัวของ TU ที่ชัดเจนขึ้นในปี 2024F ขณะที่ GFPT เติบโตจากธุรกิจไก่ในประเทศและส่งออกไก่ที่ฟื้นตัว  \nFig 1. Earnings preview for 1Q24F and 2024F  \n| Profit and Loss (THB mn) |||||||||||\n| Year-end 31 Dec | 1Q23 | 4Q23 | 1Q24F | %y-y | %q-q | 2022 | 2023 | %y-y | 2024F | %y-y |\n| - | - | - | - | - | - | - | - | - | - | - |\n| Revenue | 32,652 | 35,529 | 33,898 | 3.82 || (4.59) 155,586  | 136,153 | (11.98) | 141,157 | 3.68 |\n| Gross Profit | 4,917 | 6,327 | 5,695 | 15.8 | (10.0) | 27,206 | 23,225 | (13.8) | 24,844 | 7.0 |\n| EBITDA | 1,725 | 3,124 | 2,597 | 50.5 | (16.9) | 11,971 | 10,698 | (8.6) | 11,781 | 10.1 |\n| Interest expense | (532) | (630) | (655) | 23.0 | 3.9 | (1,998) | -2,302 | 13.3 | -2,637 | 14.5 |\n| Corporate tax | 242 | 40 | (106) | (143.9) | (364.5) | 840 | 620 | (19.0) | -616 | (199.4) |\n| Forex gain (loss) | 215 | 67 | (50) | (123.3) | (174.6) | 800 | -236 | (131.0)  :unselected: || 0 (100.0) |\n| Other extra items | 1 | -18,433 || 0 (100.0)  | (100.0) | (665) | -18,459 | Down | O :unselected: | (100.0) |\n| Gn (Ls) from Avanti+Other | 211 | 154 | 174 | (17.5) | 13.0 | 600 | 679 | 16.6 | 700 | 3.1 |\n| Gn (Ls) from Red Lobster | 20 | (554) | O | (100) | (100) | (1,630) | -1,219 | Up | O :unselected: | Up |\n| Net Profit | 1,021 | -17,189 | 1,040 | 1.8 | (106.1) | 7,138 | -13,933 | Down | 5,875 | Up |\n| Norm. Profit | 806 | 1,177 | 1,090 | 35.3 | (7.4) | 7,003 | 4,762 | (28.1) | 5,875 | 23.4 |\n| EPS (THB) | 0.21 | (3.52) | 0.21 | 1.8 | (106.1) | 1.47 | (3.15) | Down | 1.26 | Up |\n| Gross Margin (%) | 15.06 | 17.81 | 16.80 | | | 17.49 | 17.06 | | 17.60 | |\n| SG&A/Sales (%) | 12.62 | 11.74 | 12.00 | | | 12.31 | 11.98 | | 12.00 | |\n| EBITDA Margin (%) | 5.28 | 8.79 | 7.66 | | | 7.69 | 7.86 | | 8.35 | |\n| Norm profit Margin(%) | 3.13 | (48.38) | 3.07 | | | 4.50 | 3.50 | | 4.16 | |\n| Net Margin (%) | 3.13 | (48.38) | 3.07 | | | 4.59 | (10.23) | | 4.16 | |\n| Current ratio (x) :selected: | 2.14 | 1.70 | 1.72 | | | 2.38 | 1.70 | | 1.72 | |\n| Interest Coverage (x) :selected: | 1.49 | 3.42 | 2.48 | | | 4.03 | 3.00 | | 3.00 | |\n| Debt to Equity (x) :selected: | 0.70 | 1.03 | 1.01 | | | 0.68 | 1.03 | | 0.98 | |\n| BV (THB/shr) | 16.88 | 12.59 | 12.82 | | | 16.90 | 12.59 | | 13.34 | |\n| ROE (%) | 5.75 | (102.55) | 5.93 | | | 10.20 | (20.01) | | 9.73 | |  \nSource: TU and KCS estimate  \n<!-- PageFooter="IRIS Division" -->  \n<!-- PageNumber="2" -->',
 '      คู่มือพนักงาน\n\n     บริษัท เค.พี.เทรดดิ้ง  กรุ๊ป  จำกัด\n\nRevised :1 พฤษภาคม 2562\n\nการจ้างงาน\n\nการจ้างพนักงานใหม่\n\n\tการจ้างพนักงานใหม่จะกระทำเมื่อบริษัทฯมีตำแหน่งและผู้สมัครมีคุณสมบัติ ความชำนาญซึ่งสามารถทำงานอย่างมีประสิทธิผลต่อบริษัท ฯเท่านั้น โดย\n\nผู้สมัครทุกคนจะได้รับการปฏิบัติอย่างยุติธรรมและเท่าเทียมกันโดยไม่คำนึงถึงเชื่อชาติ ศาสนา ผิว หรือเพศ\n\nการคัดเลือกผู้สมัคร จะพิจารณาจากคุณสมบัติของผู้สมัครเป็นเบื้องต้น โดยไม่คำนึงถึงความสัมพันธ์ส่วนตัว การลำเอียง หรือสินบนใดๆ\n\nประเภทของพนักงาน\n\nพนักงานรายเดือน คือพนักงานที่บริษัทฯตกลงจ้างให้เป็นการประจำโดยจ่ายค่าจ้างเป็นรายเดือนอันรวมถึงค่าจ้างในวันหยุดประจำสัปดาห์ วันหยุดตามประเพณี วันหยุดพักผ่อนประจำปี และวันลาตามเงื่อนไขที่กำหนดไว้\n\nพนักงานรายวัน คือพนักงานที่บริษัทฯตกลงจ้างไว้เป็นการประจำโดยจ่ายค่าจ้างเป็นรายวัน มีสิทธิ์ได้รับค่าจ้างเฉพาะวันที่มาทำงานและได้รับค่าจ้างในวันหยุดตามประเพณี และวันหยุดพักผ่อนประจำปี และวันลาตามเงื่อนไขที่กำหนดไว้\n\nการทดลองงาน\n\n\tพนักงานใหม่ทุกคนต้องอยู่ในระยะทดลองงานไม่เกิน 120 วัน ซึ่งในระยะเวลาดังกล่าว พนักงานจะได้รับการประเมินผลงาน และความเหมาะสมโดยเปรียบเทียบกับมาตราฐานที่กำหนดไว้ภายใน 90 วัน อย่างไรก็ตาม บริษัทฯอาจเลิกจ้างพนักงานเหล่านี้ออกเมื่อใดก็ได้ระหว่างการทดลองงานหากผลการปฏิบัติงานไม่เป็นที่\n\n-1-\n\nพอใจ หรือไม่มีความเหมาะสมกับงานในหน้าที่ โดยไม่จำเป็นต้องแจ้งล่วงหน้า และไม่จ่ายเงินชดเชยให้แต่ประการใดทั้งสิ้นและในระยะทดลองงาน ห้ามขาดงาน  และถ้าป่วยต้องมีใบรับรองแพทย์เป็นหลักฐาน\n\nการเลื่อนตำแหน่ง\n\n\tบริษัทฯมีนโยบายที่จะบรรจุพนักงานปัจจุบัน สำหรับตำแหน่งว่าง ก่อนที่จะพิจารณาจากผู้สมัครภายนอก บริษัทฯจะเป็นผู้ออกประกาศอย่างเป็นทางการในการเลื่อนตำแหน่งนั้นๆ',
 'เวลาทำงานปกติ วันทำงานปกติและวันหยุดประจำสัปดาห์\n\nชั่วโมงการทำงานต่อวัน\n\nพนักงานทุกท่านต้องทำงานวันละ 8 ชั่วโมง ไม่รวมเวลาพัก\n\nการกำหนดเวลาทำงาน\n\nวันทำงานปกติ วันจันทร์-วันเสาร์\n\nเวลาทำงานปกติวันจันทร์-วันศุกร์\t08.30-17.00 น.\n\nเวลาทำงานปกติวันเสาร์\t\t08.30-12.00 น.\n\nเวลาพักกลางวัน\t\t\t12.00-13.00 น.\n\nวันหยุดตามประเพณี\n\n\tบริษัทฯ กำหนดให้มีวันหยุดตามประเพณีอย่างน้อยปีละไม่น้อยกว่า 13 วันโดยรวม โดยบริษัทฯจะประกาศวันหยุดให้ทราบล่วงหน้าเป็นประจำในเดือนธันวาคมของแต่ละปี\n\n2-\n\nการบันทึกเวลาทำงาน\n\nพนักงานทุกท่านต้องบันทึกเวลาเข้า-ออก การบันทึกเวลาทำงานของพนักงานมีดังนี้\n\nพนักงานทุกท่านจะต้องบันทึกเวลาเข้า-ออกด้วยตนเองห้ามมีการบันทึกเวลาแทนกันโดยเด็ดขาด  ยกเว้นพนักงานบางระดับ บางตำแหน่งที่บริษัทฯ กำหนดให้ไม่ต้องบันทึกเวลาเข้า-ออก\n\nเมื่อมาทำงานแล้วต้องบันทึกเวลา(ตอกบัตรเข้า,ออก)งานทุกครั้ง\n\n ในกรณีที่พนักงานลืมตอกบัตร ต้องให้หัวหน้างานเซ็นต์อนุมัติรับรองในการทำงาน มิฉะนั้นจะถือว่าขาดงาน\n\nการบันทึกเวลาแทนกัน บริษัทฯจะถือเป็นความผิดที่มีโทษทางวินัยทั้งสองฝ่าย   และ จะได้รับการตักเตือน-พักงาน-ไล่ออก  ตามลำดับ\n\nการผิดวินัยการทำงานการเข้า-ออก\n\nการเข้า-ออก ประตูบริษัท หลังจากที่พนักงานตอกบัตรทำงานแล้ว\n\nหากมีความจำเป็น ต้องขออณุญาติหัวหน้างานก่อน เพื่อไม่ให้มีผลกระทบต่อธุรกิจ\n\nการผิดวินัยการตรงต่อเวลา-มาสาย\n\nพนักงานที่บันทึกเวลาปฎิบัติงานสายเกิน 8.30 น ถือว่าสาย\n\n\tสาย 1  ครั้ง/ 1 เดือน ไม่โดนหักเงิน\n\n\tสายครั้งที่ 2 จะถูกหักงินตามชั่วโมงเต็ม\n\n\tสายมากกว่า 3 ครั้ง/เดือน หัวหน้าตักเตือน และเสียประวัติในการประเมินผล\n\nการผิดวินัยการทำงาน-การแต่งเครื่องแบบ\n\nพนักงานต้องสวมเสื้อฟอร์มที่บริษัทฯกำหนดให้ในขณะปฏิบัติงาน มิฉะนั้นถือว่าไม่พร้อมที่จะเข้าทำงานได้ (ยกเว้นวันเสาร์)\n\n3-\n\nกางเกง/ กระโปรง ต้องสุภาพ ไม่ดูอุจาดตาที่จะทำให้เสื่อมเสียภาพลักษณ์ของบริษัท\n\nรองเท้า\n\nเพื่อความปลอดภัยแก่ตัวพนักงานเอง ขณะปฏิบัติงานกับเครื่องจักรหรือขนสายไฟ พนักงานต้องสวมใส่รองเท้าหุ้มนิ้วเท้าให้มิดชิด ห้ามสวมรองเท้าชำรุดเข้าทำงาน\n\nเพื่อภาพลักษณ์ในการติดต่อลูกค้า พนักงานควรใส่รองเท้าสุภาพ ห้ามสวมรองเท้าแตะสกปรก หรือถอดรองเท้าพบปะลูกค้า หากฝ่าฝืนให้หยุดการทำงาน',
 'วินัยและความประพฤติ\n\n\tพนักงานแต่ละท่านได้รับการคาดหมายให้ปฏิบัติหน้าที่ของตนตามความสามารถอย่างดีที่สุด และต้องประพฤติปฏิบัติให้เหมาะสมเพื่อชื่อเสียงอันดีของ\n\nบริษัทฯ ทั้งในและนอกสถานที่ พนักงานทุกท่านต้องเคารพ เชื่อฟัง คำสั่งอันชอบด้วยกฏหมายของหัวหน้างาน และต้องปฏิบัติตามกฎ ตลอดจนระเบียบต่างๆ ซึ่งกำหนดไว้\n\nการดำเนินทางวินัย\n\n\tวัตถุประสงค์ของการดำเนินการทางวินัย คือการแก้ไข ปรับปรุงทัศนคติ นิสัย หรือความประพฤติของพนักงาน  บริษัท ฯ จะดำเนินการทางวินัยต่อพนักงานในกรณีของการไร้ประสิทธิภาพหรือฝ่าฝืนกฏ ระเบียบ หรือขัดขืนคำสั่งอันชอบด้วยกฏหมายดังนี้\n\nการไร้ประสิทธิภาพ\n\nในกรกณีที่บริษัทฯ มีความเห็นว่าพนักงานไม่สามารถปฏิบัติหน้าที่ของตนเองอย่างมีประสิทธิภาพ ทางบริษัทฯ อาจเลิกจ้างพนักงานผู้นั้นเมื่อใด\n\n4-\n\nก็ได้ โดยการแจ้งให้ทราบล่างหน้าอย่างน้อย 1 เดือน ภายหลังออกหนังสือเตือนในเรื่องดังกล่าว\n\nการฝ่าฝืนกฏระเบียบ บริษัท หรือขัดขืนคำสั่งอันชอบด้วยกฏหมาย\n\nการดำเนินการทางวินัยจะรุนแรงระดับใดขึ้นอยู่กับเจตนา ลักษณะของการ\n\nกระทำและผลจากการกระทำนั้นๆ โดยไม่จำเป็นต้องเรียงลำดับ และระดับต่างๆ ที่บริษัทฯจะเลือกใช้ดังนี้\n\nก. การเตือนด้วยวาจา เป็นการดำเนินการทางวินัยที่มีลักษณะเบาที่สุด\n\n                    จำนวน  …2...ครั้ง\n\n \tข. การเตือนเป็นลายลักษณ์อักษร เป็นการดำเนินการทางวินัย ในกรณีที่มี\n\nการละเมิดกฏ ข้อบังคับที่สำคัญหรือร้ายแรงหรือการกระทำความผิดวินัย จำนวนรวมครั้งที่ออกใบเตือน..2..ครั้ง   อายุความ  1  ปีนับจากวันเตือนครั้งแรก ประวัติการออกใบเตือนนี้จะนำไปพิจารณาโบนัสและเงินเดือนใน\n\nปีต่อๆไปด้วย และถ้าออกใบเตือนครั้งที่3  จะให้  พักงาน  ทันทีเป็นเวลา\n\n1 อาทิตย์ โดยไม่จ่ายค่าจ้าง\n\nค.การพักงาน เป็นการดำเนินการทางวินัยขั้นรุนแรง แต่ไม่ถึงขั้นไล่ออก ซึ่งเป็นการสั่งให้หยุดงานชั่วคราว โดยไม่จ่ายเงิน   แล้วทำผิดอีก  จะถูกให้ปลดออกโดยทันที',
 'ง. การไล่ออก เป็นการดำเนินการทางวินัยขั้นรุนแรงที่สุด โดยบริษัทฯจะไม่จ่ายเงินชดเชยใดๆ ทั้งสิ้น\n\nความผิดที่จะออกใบเตือน\n\nการทะเลาะวิวาทในสถานที่ทำการของบริษัทฯ\n\nการกระทำเจตนาทำให้ทรัพย์สินของบริษัฯ เสียหาย\n\nการกระทำโดยก่อให้เกิดอันตรายต่อตนเองและผู้อื่น\n\nการเข้าทำงานในสภาพมึนเมาจากพิษสุราหรือสิ่งเสพติดทุกชนิด\n\n5-\n\nการครอบครองหรือดื่มสุราภายในบริษัทฯโดยไม่ได้รับอนุญาต\n\nการตอกบัตรเข้างานแทนกัน\n\nการละเลยหน้าที่ ของตนเอง (อ้างอิงตามใบ Job Description)\n\nการกระทำที่ทำให้บริษัทเสียหายเช่น การกล่าวร้ายบริษัทในสื่อ Social Networkที่ทำให้บริษัทเสียภาพลักษณ์ และชื่อเสียง\n\nการตักเตือน  การตักเตือนด้วยวาจา  และการออกใบเตือน   ทุกครั้งจะต้องมีการบันทึกลงสมุด พร้อมลายเซ็นต์ ของหัวหน้าแผนก และ ฝ่ายบุคคล   \n\nความผิดที่ไล่จะออกทันที\n\n        1.การขโมย ยักยอก ทรัพย์สินของบริษัทฯ หรือของพนักงานอื่นๆ\n\n        2.การครอบครอง พกพา หรือขาย หรือเสพสิ่งเสพติดภายในบริษัทฯ\n\n        3.การละทิ้งหน้าที่เป็นเวลา 3 วัน ทำงานติดต่อกันไม่ว่าจะมีวันหยุดคั่นหรือไม่    ก็ตามโดยไม่มีเหตุอันควร\n\n        4.ฝ่าฝืนกฏข้อบังคับด้านความปลอดภัยซึ่งอาจก่อให้เกิดความเสียหายอย่างร้ายแรง\n\n       5.เจตนาปลอมแปลงหรือแก้ไขเอกสารสำคัญที่เกี่ยวข้องกับการเงินต่างๆ\n\n    **  เรื่องการเคารพต่อหัวหน้างาน ต้องเชื่อฟังหัวหน้างานและฝ่ายบุคคล ถ้าไม่เชื่อฟังจะโดนตักเตือน ทันที\n\n     **พฤติกรรมการทำงาน ห้ามใช้อารมณ์กับเพื่อนร่วมงาน เมื่อมาทำงานแล้วต้องปรับอารมณ์ให้ดี เพื่อทำงาน\n\n         ทั้ง 2 ข้อนี้ จะมีกฎระเบียบดังนี้\n\n         ครั้งแรก ตักเตือนดัวยวาจา\n\n         ครั้งที่ 2 โดนใบเตือนทันที\n\n         ครั้งที่ 3 สั่งพักงานและ ตัดเงิน\n\n         ครั้งที่ 4  ให้ออก\n\n6-',
 '   ***กรณีทะเลาะวิวาทรุนแรง  ***บริษัทฯสามารถใล่ออกจากงานทันที  \n\nเงินเดือน\n\nวันและสถานที่จ่ายค่าจ้าง\n\n\tบริษัทฯ จะจ่ายค่าจ้างแก่พนักงานเดือนละ 1 ครั้งทุกสิ้นเดือน โดยจ่ายเป็นเงินสด ณ ที่ทำการของบริษัทฯสำหรับพนักงานที่ยังไม่ครบทดลองงาน  และพนักงานที่ครบทดลองงานแล้วจะจ่ายผ่านธนาคารกรุงเทพฯ\n\nการปรับค่าจ้างและเงินเดือน\n\n\tบริษัทฯได้กำหนดให้มีการปรับค่าจ้างและเงินเดือน 3 ลักษณะ ดังนี้คือ\n\nการปรับค่าจ้างตามกฏหมายบังคับ\n\nการปรับประจำปี โดยปกติการปรับเงินเดือนลักษณะนี้จะกระทำปีละ 1 ครั้งโดยมีผลจากวันที่ 1 มกราคม ของแต่ละปี อย่างไรก็ตาม การปรับเงินเดือนลักษณะนี่ มิใช่  การปรับโดยอัตโนมัติ แต่จะพิจารณาจากผลการปฏิบัติงานของพนักงาน (ผลการทำงานโดยทั่วไป ความประพฤติ ผลสำเร็จจากการทำงาน)\n\nการปรับเมื่อเลื่อนตำแหน่ง การปรับเงินเดือนลักษณะนี้จะกระทำเมื่อพนักงานได้รับการเลื่อนตำแหน่งที่สูงขึ้นและสามารถปฏิบัติหน้าที่ในตำแหน่งใหม่ได้เป็นอย่างดี หรือในกรณีที่พนักงานได้รับการประเมิณผลการปฏิบัติงานในระดับยอดเยี่ยม\n\nค่าล่วงเวลาและค่าทำงานในวันหยุด\n\n\tการทำงานล่วงเวลา คือการทำงานที่ได้รับอนุมัติให้ทำเกินเวลาทำงานปกติของพนักงานในวันทำงาน หรือในวันหยุดประจำสัปดาห์ หรือในวันหยุดตามประเพณี\n\nที่บริษัทฯประกาศ     หัวหน้างาน ต้องเป็นผู้อนุมัติเป็นลายลักษณ์อักษรก่อนที่จะ\n\n7-\n\nอนุญาตให้มีการทำงานล่วงเวลา  หรือทำงานในวันหยุดยกเว้นในกรณีฉุกเฉิน',
 'หลักเกณฑ์การจ่ายค่าล่วงเวลาและค่าทำงานในวันหยุด\n\nการทำงานล่วงเวลาให้แยกระหว่างแผนกขนส่ง กับ  แผนกสต๊อกสินค้า\n\nพนักงานที่ได้รับมอบหมายให้ทำงานล่วงเวลาในวันทำงานปกติ จะได้รับค่าล่วงเวลาในอัตรา 1.5 เท่าของอัตราค่าจ้างต่อชั่วโมงในเวลาทำงานปกติ    แผนกขนส่งเริ่มจากเวลา 17.00  แผนกสต๊อคเริ่มจากเวลา17.30 น. เป็นต้นไป (สต๊อคต้องลงเวลาทำงานจริงเท่านั้น)\n\nพนักงานที่ได้รับมอบหมายให้ทำงานล่วงเวลาในการทำงานวันหยุด คือวันหยุดประจำสัปดาห์ หรือวันหยุดตามประเพณีของบริษัทฯ (หลังเวลา 17.30น.) จะได้รับค่าล่วงเวลาในอัตรา 2 เท่าของอัตราค่าจ้างต่อชั่วโมงในเวลาทำงานปกติ\n\nพนักงานที่ทำงานล่วงเวลาต้องบันทึกการทำงานล่วงเวลาในแบบฟอร์มที่บริษัทฯ กำหนดไว้เพิ่มเติม และให้หัวหน้างานเซ็นรับรองการทำงานก่อนส่งให้ฝ่ายบุคคล\n\nการทำงานในวันหยุดประจำสัปดาห์และวันหยุดตามประเพณี\n\nพนักงานรายเดือนซึ่งได้รับค่าจ้างในวันหยุดอยู่แล้ว หากได้รับมอบหมายให้ทำงานวันหยุดประจำสัปดาห์หรือวันหยุดตามประเพณีจะได้รับค่าจ้างเพิ่มไม่เกิน 1 เท่า ของค่าจ้างปกติ\n\nพนักงานรายวัน ซึ่งได้รับมอบหมายให้มาปฏิบัติงานในวันหยุดประจำประจำสัปดาห์จะได้รับค่าจ้างเพิ่มไม่เกิน 2 เท่าของอัตราค่าจ้างปกติ\n\nพนักงานรายวัน ซึ่งได้รับมอบหมายให้มาปฏิบัติงานในวันหยุดตามประเพณี จะได้รับค่าจ้างเพิ่มไม่เกิน 1 เท่าของอัตราค่าจ้างปกติ เพราะพนักงานได้รับค่าจ้างในวันหยุดตามประเพณีอยู่แล้ว\n\n8-\n\nการลาหยุด\n\nวันลา\n\n\tบริษัทฯอนุญาตให้พนักงานลาหยุดได้ตามประเภทของการลาต่างๆ กัน พนักงานต้องมีความรับผิดชอบต่อบริษัทฯ และเพื่อนร่วมงาน โดยลาเฉพาะเหตุที่จำเป็นเท่านั้น ซึ่งการหยุดงานของท่านนั้น อาจสร้างความยุ่งยาก หรือความลำบากให้กับเพื่อนร่วมงานได้ พนักงานต้องกรอก “ใบลา” สำหรับการลาทุกประเภท และมอบต่อฝ่ายบุคคลเพื่ออนุมัติทุกครั้งที่ลาหยุด ในบางกรณีพนักงานต้องแนบเอกสารในการลาด้วย',
 '\tพนักงานทุกท่านซึ่งปฏิบัติงานมาครบ 1ปีจะได้รับวันหยุดประจำปี ดังนี้\n\nวันหยุดพักผ่อนประจำปี\n\nพนักงานทุกท่านจะได้รับวันหยุดพักผ่อนประจำปีท่านละ  10 วัน สำหรับพนักงานใหม่ที่ผ่านการทดลองงานแล้วก็ให้ลดลงตามส่วน\n\nถ้าลาเกินสิทธิ 10 วัน  ถือว่าเป็นการหยุดงานไม่รับค่าจ้าง หักตามจำนวนวันที่หยุดเกินตามค่าจ้างรายวัน\n\nการลาป่วย\n\nพนักงานมีสิทธิลาป่วยได้ตามจำนวนที่ป่วยจริง บริษัทฯจะจ่ายค่าจ้างให้ไม่เกิน 30 วันทำงาน แต่ต้องมีใบรับรองแพทย์จากโรงพยาบาลที่ทำประกันสังคมเท่านั้น (ไม่ขอรับใบรับรองแพทย์จากทางคลีนิค) และให้โทรลาแจ้งหัวหน้างานได้  และต้องส่งใบลาพร้อมแนบใบรับรองแพทย์ในวันที่กลับมาทำงานให้หัวหน้างานเซ็นต์รับรอง และนำส่งทางบัญชีบุคคล มิฉะนั้นจะถือเป็นการขาดงาน\n\nกรณีการไปพบแพทย์ตามนัดถือเป็นการลาป่วย(ให้แนบใบนัด)\n\nพนักงานที่แจ้งลาป่วยเป็นเท็จจะต้องถูกดำเนินการทางวินัยตามระเบียบ\n\n  ของบริษัทฯ ด้วย\n\n9-\n\nวันลาเมตตาจิตที่ได้รับค่าจ้างเท่ากับวันทำงานปกติ\n\nไม่ถือเป็นการลากิจหรือลาป่วย มีดังต่อไปนี้\n\n3.1 วันลาถึงแก่กรรมของครอบครัว\n\nบิดา มารดา คู่สมรสและบุตรที่ถูกต้องตามกฏหมาย\n\nบริษัทให้สิทธิ์ลาหยุดได้ 7 วัน โดยต้องแนบใบมรณบัตรด้วย\n\n3.2 วันลาคลอด\n\nสตรีที่มีครรภ์ ลาก่อนและหลังคลอดได้ไม่เกิน 90 วัน รวมวันนัดหมอตรวจครรภ์พนักงานจะได้รับค่าจ้างระหว่างลาคลอดไม่เกิน 45 วัน\n\nสามี มีสิทธิลาเมื่อภรรยาคลอดบุตรได้ 5 วัน\n\n3.3 วันลาอุปสมบท\n\nเป็นการลาได้เพียงครั้งเดียว ไม่เกิน 15 วัน (นับรวมวันหยุด) โดยพนักงานต้องทำงานครบ 1 ปี ถึงจะได้รับสิทธิ์ได้\n\nการลากิจ (ไม่ได้รับค่าจ้าง)\n\nการลากิจกรณีต่างๆ พนักงานต้องยื่นใบลาก่อนล่วงหน้า 1 วัน (กรณีมีเหตุพิเศษไม่อาจส่งใบลาล่วงหน้าได้ ให้โทรศัพท์แจ้งหัวหน้างาน และจัดส่งใบลาทันทีในวันแรกที่มาปฏิบัติงานให้กับหัวหน้างานนำส่งให้บัญชีบุคคล)\n\nเขียนใบลาและส่งใบลาในกรณีต่างๆ\n\nถือเป็นหน้าที่ของตัวพนักงานที่จะต้องส่งใบลาล่วงหน้า ให้กับหัวหน้างาน อนุญาติ รับรองการลาว่าสามารถจัดการได้ไม่กระทบก่อความเสียหายให้บริษัท นำใบลาที่หัวหน้าอนุมัติส่งให้ฝ่ายบัญชีบุคคล หากไม่ปฎิบัติตามจะ',
 'ถือว่าเป็นการผิดวินัยการลา บกพร่องในการทำงาน เป็นการขาดงานละทิ้งหน้าที่โดยไม่มีเหตุอันสมควร\n\n10-\n\nพนักงานที่ไม่มีสิทธิได้รับเงินค่าล่วงเวลาหรือค่าทำงานในวันหยุด\n\nพนักงานที่สมัครใจทำงานล่วงเวลา หรือทำงานในวันหยุด โดยหัวหน้างานมิได้มอบหมายและอนุญาต\n\nพนักงานที่หัวหน้างานไม่ได้เซ็นรับรองการทำงานล่วงเวลาหรือทำงานวันหยุดในแบบฟอร์มการบันทึก\n\nการพ้นสภาพการเป็นพนักงาน\n\n\t1.  การลาออก\n\n\tพนักงานทุกคนซึ่งครบระยะทดลองงานแล้ว และประสงค์จะลาออกต้องแจ้ง\tให้บริษัทฯ ทราบล่วงหน้าเป็นลายลักษณ์อักษร อย่างน้อย 1 เดือน\n\n\t2.  การปลดออก\n\n\tหมายถึงการที่บริษัทฯ ให้พนักงานพ้นสภาพจากการเป็นพนักงานด้วยสาเหตุดังต่อไปนี้\n\nการลดอัตรากำลัง พนักงานจะถูกเลิกจ้างเพื่อเป็นการลดอัตรากำลัง ในกรณีที่บริษัทฯมีความจำเป็นในสถานการณ์ต่างๆ การเลิกจ้าง เพื่อลดอัตรากำลังนี้มิได้เกิดจากความผิดของพนักงานแต่เป็นผลจากการเปลี่ยนแปลงทางด้านเทคโนโลยีการจัดรูปแบบองค์การใหม่ ความตกต่ำทางเศรษฐกิจ หรือการเงินของบริษัทฯ เป็นต้นโดยบริษัทจะจ่ายค่าชดเชยรายได้ให้\n\nถูกสั่งพักงาน ครบ 1  ครั้งแล้ว  และยังทำผิดซ้ำอีก\n\n\t3.  การไล่ออก\n\n\t เนื่องจากพนักงานกระทำผิดระเบียบ วินัยร้ายแรง ซึ่งมีเหตุผลพอเพียงที่จะไล่ออกได้ทันทีที่กระทำผิด เป็นผลจากการลงโทษทางวินัยตามขั้นตอนที่ระบุไว้ในกฏระเบียบของบริษัทฯพนักงานผู้นั้นมีความผิด ดังต่อไปนี้\n\n11-\n\n1. ทุจริตต่อหน้าที่หรือกระทำความผิดอาญาโดยเจตนาแก่นายจ้าง\n\n2 .จงใจทำให้นายจ้างได้รับความเสียหาย\n\n3.ประมาทเลินเล่อเป็นเหตุให้นายจ้างได้รับความเสียหายอย่างร้ายแรง\n\n4. ฝ่าฝืนข้อบังคับเกี่ยวกับการทำงาน ระเบียบ หรือคำสั่งของนายจ้างอัน ชอบด้วยกฎหมายและเป็นธรรม และนายจ้างได้ตักเตือนเป็นหนังสือแล้ว\n\n เว้นแต่กรณีที่ร้ายแรงนายจ้างไม่จำเป็นต้องตักเตือน ทั้งนี้ หนังสือเตือน',
 'ให้มีผลบังคับได้ไม่เกิน 1 ปี นับแต่วันที่ลูกจ้างได้กระทำผิด   \n\n5.ละทิ้งหน้าที่เป็นเวลา 3 วันทำงานติดต่อกันโดยไม่มีเหตุผลที่สมควร\n\n6.ได้รับโทษจำคุกตามคำพิพากษาถึงที่สุดให้จำคุก\n\n\t4.  การเกษียณอายุ\n\n\tอายุเกษียณตามปกติของพนักงานทุกคน คือ 55 ปีบริบูรณ์ อย่างไรก็ตามพนักงานอาจสมัครใจของเกษียณอายุก่อนกำหนดได้ หากพนักงานทำงานมาแล้วไม่น้อยกว่า 15ปี และมีอายุไม่น้อยกว่า 55 ปี บริษัทฯของสงวนสิทธิ์ที่จะจ้างพนักงานที่ครบเกษียณแล้วหรือปฏิเสธคำขอสมัครใจเกษียณอายุก่อนกำหนดก็ได้\n\nแนวปฏิบัติในการพ้นสภาพการเป็นพนักงาน\n\nการเลิกจ้างมีดังนี้\n\nคืนทรัพย์สินทุกอย่างของบริษัทฯ ทรัพย์สินต่างๆนั้น รวมถึงทรัพย์สินที่เป็นรูปธรรมเช่น คอมพิวเตอร์ เครื่องมือ อุปกรณ์ต่างๆ และทรัพย์สินที่ไม่เป็นตัวตน เช่นเอกสารโต้ตอบต่างๆ เป็นต้น ข้อที่ควรเอาใจใส่ในการคืนทรัพย์สินต่างๆ นั้นต้องให้บริษัทฯ รับทราบด้วย\n\n12-\n\nสวัสดิการและอื่นๆ\n\nโบนัสประจำปี\n\n                บริษัทฯมีโบนัสให้กับพนักงานทุกปีเพื่อเป็นขวัญและกำลังใจให้กับพนักงาน\n\nกองทุนประกันสังคม\n\n\tกองทุนประกันสังคมได้รับการจัดตั้งโดยกระทรวงแรงงาน และสวัสดิการทางสังคม เพื่อช่วยเหลือทางการเงินด้านค่ารักษาพยาบาล หรือการขาดรายได้ อันเป็นผลมาจากการเจ็บป่วย อุบัติเหตุ ทุพพลภาพ หรือเสียชีวิต ซึ่งไม่เกี่ยวกับหน้าที่การงานของพนักงาน และรวมถึงการคลอดบุตรด้วย                                                        \n\n\tกองทุนนี้เป็นการจ่ายเงินสมทบร่วมกันในอัตราเท่ากันทุกเดือนระหว่าง\n\nบริษัทฯ และพนักงาน อัตราการจ่ายสมทบปัจจุบันคือ ร้อยละ 5 ซึ่งคำนวณจากเงินเดือนของพนักงานและส่วนบริษัทฯ จะต้องจ่ายสมทบอีก 5 % ของยอดเงินหักทั้งหมด  \n\nการตรวจสุขภาพประจำปี\n\n                บริษัทฯได้จัดให้มีการตรวจสุขภาพประจำปีให้กับพนักงานทุกคน',
 'เครื่องแบบ\n\n\tเพื่อเป็นการช่วยเหลือค่าใช้จ่ายในการแต่งกายของพนักงานและเพื่อให้การแต่งกายเป็นไปในลักษณะเดียวกัน ทางบริษัทฯ จึงเป็นผู้จัดสรรเสื้อฟอร์มให้กับพนักงานคนละ 4 ตัว ต่อปี\n\nการจัดงานเลี้ยงสังสรรค์ปีใหม่\n\n                บริษัทฯได้จัดให้มีงานเลี้ยงสังสรรค์ปีใหม่ให้กับพนักงานทุกคน\n\n13-\n\nความสำคัญของพนักงานที่มีต่อบริษัทฯ\n\n\tวัตถุประสงค์สำคัญในการบริหารงานบุคคลของบริษัทฯ คือการว่าจ้าง การจูงใจ การรักษาไว้ และการพัฒนาคุณภาพ ของพนักงานทุกคนเพื่อให้บรรลุเป้าหมายของบริษัทฯ อย่างมีประสิทธิภาพ โดยจ่ายค่าตอบแทนและผลประโยชน์ที่เป็นธรรมแก่พนักงาน ทั้งนี้ขึ้นอยู่กับขอบเขตหน้าที่และความรับผิดชอบของแต่ละตำแหน่ง\n\nนโยบายทั่วไปในการบริหารงานบุคคลของบริษัทฯมีดังนี้\n\nการรักษาและส่งเสริมให้เกิดความเข้าใจและความสัมพันธ์อันดีระหว่างพนักงานกับผู้บริหารตลอดจนระหว่างพนักงานด้วยกัน\n\nจัดให้มีการประเมินผลอย่างมีประสิทธิภาพ และเป็นธรรมแก่พนักงาน โดยการเปรียบเทียบกับหน้าที่ความรับผิดชอบ และเป้าหมายที่กำหนดไว้ล่วงหน้าอย่างน้อยปีละครั้ง\n\nปรับปรุงและรักษาสภาพการทำงาน และเครื่องมือ อุปกรณ์ต่างๆ ให้ถูกต้องตามมาตรฐานความปลอดภัยและสุขอนามัย\n\nส่งเสริมให้พนักงานทุกท่านได้รับการพิจารณาเลื่อนตำแหน่งโดยเท่าเทียมกัน\n\nพิจารณาแก้ไขข้อร้องทุกข์ของพนักงานโดยไม่ชักช้า\n\nสำรวจทัศนคติของพนักงาน ที่มี ต่อนโยบาย ข้อบังคับ และสภาพการทำงานของบริษัทฯอย่างสม่ำเสมอ\n\n               กฎการใช้มือถือของบริษัทฯ\n\n           \t1.ห้ามใช้คุยเรื่องส่วนตัว\n\n \t2.ห้ามผู้ไม่มีกิจธุระ เอาโทรศัพท์มือถือบริษัทฯ  กลับบ้าน นอกจากผู้นั้นได้รับ อนุญาตแล้วเท่านั้น    \n\n14-',
 'ความคาดหวังของบริษัทฯที่มีต่อพนักงาน\n\nพนักงานทุกคนต้องปฏิบัติตามกฏ และข้อบังคับในการทำงาน ซึ่งชอบด้วยกฏหมายที่บริษัทฯ กำหนด และเห็นชอบร่วมกันแล้ว ตลอดจนต้องยอมรับสิทธิของฝ่ายบริหารในการดำเนินธุรกิจให้เป็นไปตามนโยบาย เป้าหมาย และมาตรฐานซึ่งได้กำหนดไว้โดยไม่กระทบต่อสิทธิและผลประโยชน์ของพนักงาน\n\nพนักงานแต่ละคนต้องทราบหน้าที่และความรับผิดชอบตลอดจนเงื่อนไขการว่าจ้างของตนเองและต้องปฏิบัติตามอย่างเคร่งครัด\n\nพนักงานทุกคนต้องทำงานร่วมกันอย่างเต็มกำลังความสามารถ และระลึกเสมอว่าความสัมพันธ์ในการทำงาน ทัศนคติ และความประพฤติของพนักงานทุกคนจะก่อให้เกิดผลกระทบต่อเพื่อนร่วมงาน และบริษัทฯซึ่งจะรวมไปถึงผลผลิต กำไร ค่าจ้าง เงินได้อื่นๆ ตลอดจนความเป็นอยู่ของ\n\nพนักงานด้วย ดังนั้นพนักงานต้องให้ความร่วมมือสนับสนุนฝ่ายบริหารในการพัฒนาปรับปรุงการผลิต ประหยัดค่าใช้จ่าย และปกป้องผลประโยชน์ของบริษัทฯ\n\nพนักงานทุกคนต้องรับผิดชอบในการดูแลรักษา อาคาร สถานที่ เครื่องใช้ และอุปกรณ์ต่างๆ ให้อยู่นในสภาพที่ปลอดภัย และถูกสุขอนามัย พนักงานต้องรายงานสิ่งที่อาจก่อให้เกิด การบาดเจ็บหรือมีสภาพที่ไม่น่าปลอดภัยแก่หัวหน้างานทราบทันที นอกจากนั้นพนักงานต้องใช้อุปกรณ์ป้องกันอันตรายส่วนบุคคลที่บริษัทฯ จัดหาให้ทุกครั้งสำหรับงานที่จำเป็นต้องใช้อุปกรณ์ดังกล่าว ห้ามนำออกนอกบริษัทฯ ถ้ามีเหตุจำเป็นต้องใช้ให้ขอ\n\nอนุญาติก่อน\n\n15-\n\nพนักงานทุกคนต้องพัฒนาความรู้และความสามารถในการปฏิบัติงานเพื่อให้ทันกับการเปลี่ยนแปลงต่างๆ ตลอดเวลา  ในการปฎิบัติงาน หรือว่ากระบวนการทำงานใดที่เยินเย่อไม่จำเป็น มีจุดใดที่สามารถนำเทคโนโลยีมาใช้แทนได้บ้างควรศึกษาเพิ่มเติม\n\nพนักงานต้องไม่เปิดเผยข้อมูลความลับที่รับทราบ จากการปฏิบัติงานในหน้าที่ของตนเอง ยกเว้นในกรณีที่ได้รับอนุญาตแล้ว และห้ามนำข้อมูลดังกล่าวไปแสวงหาผลประโยชน์ด้วยตนเอง หรือผ่านบุคคลที่สามโดยผิดจรรยาบรรณ หรือผิดกฏหมาย\n\nพนักงานต้องหลีกเลี่ยงการกระทำใดๆ ที่ขัดกับผลประโยชน์ของบริษัทฯ หรือที่ก่อให้เกิดผลเสียต่อหน้าที่การงานของตนเอง\n\nพนักงานจะต้องละเว้น การกระทำที่ผิดกฏระเบียบ หรือจรรยาบรรณของ',
 'บริษัทฯ ทั้งทางตรงและทางอ้อม\n\nมีความเข้าใจทางด้านความปลอดภัยในการปฏิบัติงานทุกขั้นตอน\n\n16-\n\nสารบัญ\n\n\t\t\t\t\t   หน้า\n\n\tการจ้างงาน\t\t\t\t\t1\n\n\tวินัยและความประพฤติ\t\t\t\t4\n\n\tเงินเดือน\t\t\t\t\t7\n\n\tการลาหยุด\t\t\t\t\t9\n\n\tการพ้นสภาพการเป็นพนักงาน\t\t\t11\n\n\tสวัสดิการและอื่นๆ\t\t\t\t13\n\nบริษัท เค.พี.เทรดดิ้ง กรุ๊ป จำกัด\n\nเป็นบริษัท ขายสายไฟและอุปกรณ์ มากมายหลายชนิดรวมทั้ง\n\nจัดทำการแปรรูปสายไฟตามที่ลูกค้าต้องการ สั่งแล้วคอยรับได้ทันที\n\n\tเราบริการจัดส่งสินค้าให้ลูกค้าได้รับของภายใน 2 วัน\n\n    ดังนั้น งานรับ Order จากลูกค้า, งานแปรรูปสินค้าและการจัดส่งสินค้าให้ทันตามที่ลูกค้าต้องการ จึงเป็น Core business ของบริษัทนี้\n\nบริษัทถือหลักการที่ว่า จะปฏิบัติตามกฎหมายคุ้มครองแรงงาน เพื่อให้ลูกจ้างทำงานด้วยความปลอดภัย มีสุขอนามัยที่ดี ได้รับค่าตอบแทนและสวัสดิการตามกฎหมายคุ้มครองแรงงานปัจจุบัน (ธ.ค. 2561)',
 '<figure>  \n![](figures/0)  \n<!-- FigureContent="Yuanta Securities" -->  \n</figure>  \nYUANTA ESG RATING  \n<figure>  \n![](figures/1)  \n<!-- FigureContent="AA" -->  \n</figure>',
 '# Thailand : Food & Beverage  \nCompany Update BUY (Upgrade)  \nTP upside (downside) +15.9%  \nClose Mar 29, 2024 Price (THB) 14.50 12M Target (THB) 16.80 Previous Target (THB) 16.80',
 "# What's new?  \nคาดกำไรปกติ 1Q67 ที่ 949 ลบ. (-25.3% QOQ, +7.3% YOY) มองเป็นระดับต่ำสุดของ ปี โดยกำไรชะลอลง QoQ จากธุรกิจ Frozen และ Pet Care ที่ลดลงตามฤดูกาล แต่ เติบโต YOY จากฐานต่ำ)  \nแนวโน้มกำไรปกติ 2Q67 คาดกลับมาเติบโต ทั้ง QOQ และ Yoy จากปริมาณขายที่เติบโต และบริษัทจะรับรู้ราคาต้นทุนวัตถุดิบที่ ลดลงได้มากขึ้นหนุน GPM",
 '## Our view  \nเรามีมุมมองบวกต่อแนวโน้มผลประกอบการ ของ TU ที่คาดจะเติบโตได้ทั้ง QoQ และ YOY ต่อเนื่องใน 2Q-3Q67 จากธุรกิจหลักที่ฟื้นตัว และไม่ต้องรับรู้ขาดทุนจาก Red Lobster  \nเราคงประมาณการกำไรปกติปี 2567 ที่ 5,965 ลบ. (+5.8% Yoy) คงราคาเป้าหมายที่ 16.80 บาท ราคาปัจจุบันมี Upside gain 15.9% เริ่ม น่าสนใจ จึงปรับคำแนะนำขึ้นเป็น "ซื้อ"  \nCompany profile: The Company engages in the manufacture and export of frozen and canned seafood. Its comprehensive business is completed with snack foods such as canned food, frozen food and snacks of various types especially seafood. Also, its businesses include packaging and publishing business, domestic market business, animal feed and development of shrimp species for sale business and commercial shrimp hatchery and nursery.  \nResearch Analysts:  \n<figure>  \n![](figures/2)  \n</figure>  \nKritawit Ratanakhanokchai  \n|||\n| - | - |\n| Tel. | +662 009 8064 |\n| E-Mail | Kritawit.r@yuanta.co.th |\n| ID | 124392 |  \n| Bloomberg code | TU TB |\n| - | - |\n| Market cap (THBmn) | 67,499.42 |\n| Corporate Gov.Rating | |\n| 6M avg. daily turnover (THBmn) | 296.43 |\n| Outstanding shares (mn) | 4,655.13 |\n| Free float (%) | 69.88% |\n| Major shareholders (%) | |\n| Thai NVDR Company Limited | 7.45 |\n| MR. THIRAPHONG CHANSIRI | 6.86 |\n| MITSUBISHI UFJ MORGAN STANLEY SECURITIES CO.,LTD. | 5.13 |',
 '# Financial outlook (THBmn)  \n| Year to Dec | 2022A | 2023A | 2024F | 2025F |\n| - | - | - | - | - |\n| Revenue | 155,586 | 136,153 | 142,010 | 153,083 |\n| EBITDA | 12,885 | 12,311 | 14,132 | 15,318 |\n| Core Profit | 7,201 | 5,640 | 5,965 | 6,910 |\n| Net Profit | 7,138 | -13,933 | 5,965 | 6,910 |\n| Core EPS | 1.51 | 1.21 | 1.28 | 1.48 |\n| Core EPS Growth (%) | -10.8 | -19.7 | 5.8 | 15.8 |\n| DPS | 0.84 | 0.54 | 0.73 | 0.85 |\n| Core P/E (x) | 9.6 | 12.0 | 11.3 | 9.8 |\n| P/BV(x) | 0.9 | 1.2 | 1.1 | 1.1 |\n| Div Yield (%) | 5.8 | 3.7 | 5.0 | 5.8 |\n| Net gearing (%) | 53.1 | 79.3 | 77.9 | 75.4 |\n| ROE (%) | 8.1 | -21.1 | 8.6 | 9.5 |  \nSource: Company, Yuanta Research  \n1 April 2024  \nValuation Methodology: SOTP; Seafood Business (PER 12.4x), Pet Care Business (Holding Discount 40%)',
 '# Thai Union Group (TU)  \n## คาดกำไร 1Q67 จะเป็นระดับต่ำสุดของปีนี้  \nคาดกำไรปกติ 1Q67 เติบโต Yoy ดีกว่าเดิมที่ประเมินว่าทำได้เพียงทรงตัว  \nเราคาดกำไรปกติ 1Q67 ของ TU ที่ 949 ลบ. (-25.3% QoQ, +7.3% Yoy) ดีกว่าเดิมที่เราประเมินว่าจะทำ ได้เพียงทรงตัว Yoy และอาจเป็นระดับกำไรต่ำสุดของปี โดยเราคาดรายได้ที่ 3.33 หมื่นลบ. (-6.2% QoQ, +2.1% Yoy) การชะลอลง QoQ มาจากรายได้ที่ลดลงในธุรกิจ Frozen และ Petcare จาก Demand ที่ ชะลอตัวลงในตลาดสหรัฐฯ หลังมีการสั่งซื้อไปมากใน 4Q67 กระทบต่อเนื่องต่อ GPM ให้ลดลง (ปริมาณ ขายลดลงส่งผลให้ต้นทุนต่อหน่วยสูงขึ้น) ส่วนรายได้ธุรกิจ Ambient เห็นการฟื้นตัวได้เล็กน้อย QoQ ขณะที่ต้นทุนราคาปลาทูน่าที่ปรับลดลงมาเฉลี่ยอยู่ที่ 1,350 USD/ตัน จากเฉลี่ย 1,500 USD/ตัน ใน 4Q66 จะยังไม่ได้ส่งผลบวกต่อ GPM ของ TU เนื่องจากบริษัทต้องปรับลดราคาขายสินค้ากลุ่ม OEM ตามราคา ต้นทุนทูน่า ทำให้เราคาด GPM อยู่ที่ 16.8% ลดลงจาก 17.8% ใน 4Q66 แต่เทียบ Yoy ดีขึ้นจาก 15.1% ใน 1Q66 นอกจากนี้ TU จะไม่ต้องรับรู้ส่วนแบ่งกำไร/ขาดทุนจาก Red Lobster เป็นไตรมาสแรก  \n### คาดกำไรปกติกลับมาเติบโตได้ทั้ง QoQ และ Yoy ได้ใน 2Q-3Q67  \nเราประเมินว่ากำไรของ TU จะกลับมาเติบโตทั้ง QoQ และ Yoy อีกครั้งได้ใน 2Q67 และต่อเนื่องถึง 3Q67 ที่เป็นช่วง High Season ของธุรกิจ จากการฟื้นตัวของรายได้ และ GPM แบบ QOQ อย่างค่อยเป็นค่อยไป เนื่องจากบริษัทจะเริ่มรับรู้ราคาต้นทุนเฉลี่ยที่ปรับลงได้มากขึ้น (ปกติจะมี Lag time ราว 2 - 3 เดือน) ขณะที่การปรับราคาสินค้ากลุ่ม OEM ลงใกล้สิ้นสุดแล้วตามราคาปลาทูน่าที่คาดเริ่มมี Downside จำกัด ขณะที่ฝั่ง Demand จากลูกค้าฟื้นตัวหลังมีการชะลอไปใน 1Q67 และคาดกำไรปกติจะเติบโตเด่นเทียบ YOY จากฐานต่ำในปีก่อนที่ลูกค้าตลาดสหรัฐฯ และยุโรปอยู่ในช่วงระบายสต็อกสินค้า นอกจากนี้จะไม่ต้อง รับรู้ส่วนแบ่งขาดทุนจาก Red Lobster เหมือนในปีก่อน',
 '## คงราคาเหมาะสมที่ 16.80 บาท แต่ปรับคำแนะนำขึ้นเป็น "ซื้อ"  \nเราคงมุมมองบวกต่อแนวโน้มผลประกอบการในปี 2567 ของ TU ที่คาดจะฟื้นตัวอย่างค่อยเป็นค่อยไป คาดกำไรปกติที่ 5,965 ลบ. (+5.8% Yoy) เราคงราคาเหมาะสมที่ 16.80 บาท (อิงวิธี SOTP) มี Upside gain 15.9% และราคาปัจจุบันซื้อขายบน PER67 เพียง 12 เท่า มองว่าเริ่มน่าดึงดูด จึงปรับคำแนะนำขึ้น จาก TRADING เป็น "ซื้อ" เหมาะสำหรับการลงทุนระยะยาว  \nบริษัทหลักทรัพย์ หยวนต้า (ประเทศไทย) จำกัด อาจเป็นผู้ออกและเสนอขายใบสำคัญแสดงสิทธิ์อนุพันธ์ (Derivative Warrants) ที่อ้างอิงหลักทรัพย์ใน SET100 และ SET50 index และอาจเป็นผู้ดูแลสภาพคล่อง (Market Maker) โดยบริษัทฯ อาจจัดทำบทวิเคราะห์ของหลักทรัพย์อ้างอิงดังกล่าว ดังนั้น นักลงทุนควรศึกษารายละเอียดในหนังสือชี้ชวน ของใบสำคัญแสดงสิทธิอนุพันธ์ดังกล่าวก่อนการตัดสินใจลงทุน\n<figure>  \n![](figures/3)  \n<!-- FigureContent="Yuanta Securities" -->  \n</figure>  \nWe Create Fortune\n===  \nFigure 1: Quarterly Earnings Preview  \n| (Unit : THB mn) | 1Q24F | 4Q23 | %QoQ | 1Q23 | %YoY |\n| - | - | - | - | - | - |\n| Core revenue | 33,337 | 35,529 | -6.2% | 32,652 | 2.1% |\n| COGS | (27,752) | (29,201) | -5.0% | (27,735) | 0.1% |\n| Gross profit | 5,585 | 6,327 | -11.7% | 4,917 | 13.6% |\n| SG&A | (4,048) | (4,173) | -3.0% | (4,121) | -1.8% |\n| Operating Profit | 1,884 | 2,061 | -8.6% | 1,450 | 29.9% |\n| Depre & amortization | 1,190 | 1,073 | 10.9% | 1,049 | 13.5% |\n| Operating EBITDA | 3,452 | 3,134 | n.a. | 2,499 | n.a. |\n| Other incomes (expenses) | 126 | 304 | n.a. | 424 | n.a. |\n| Financial cost | (659) | (630) | 4.6% | (532) | 23.8% |\n| Share of profit (loss) from associates | 220 | (398) | n.a. | 231 | n.a. |\n| Pre-tax profit | 1,225 | 1,431 | -14.4% | 917 | 33.5% |\n| Taxes | (66) | 40 | -264.9% | 242 | -127.4% |\n| Minorities | (283) | (244) | 16.1% | (137) | 105.9% |\n| Net profit | 875 | (18,406) | n.a. | 1,022 | n.a. |\n| Extraordinary items | (74) | (19,676) | n.a. | 137 | n.a. |\n| Core profit | 949 | 1,270 | -25.3% | 884 | 7.3% |\n| Core EPS (THB) | 0.20 | 0.27 | -25.3% | 0.19 | 7.3% |\n| Margins (%) | | | Dif QoQ | | Dif YoY |\n| Gross profit margin | 16.8 | 17.8 | -1.06 | 15.1 | 1.70 |\n| Operating margin | 5.6 | 5.8 | -0.15 | 4.4 | 1.21 |\n| EBITDA margin | 10.4 | 8.8 | 1.53 | 7.7 | 2.70 |\n| Norm profit margin | 2.8 | 3.6 | -0.73 | 2.7 | 0.14 |',
 '## Source: Company, Yuanta Research  \nFigure 2: Skipjack Tuna price  \n<figure>  \n![](figures/4)  \n<!-- FigureContent="Skipjack Tuna Price Unit : USD/Ton 2,200 2,000 1,980 2,000 2,000 2,000 1,900 1,800 1,800 1,800 1,680 1,700 1,600 1,600 1,500 1400 1,450 1,400 1300 1,200 1,000 800 Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec -2020 -- 2021 -2022 -- 2023 -2024" -->  \n<figcaption>  \nSource: Company, Yuanta Research  \n</figcaption>  \n</figure>  \n<!-- PageFooter="Thailand" -->  \n<!-- PageNumber="Page 2 of 8" -->'
 ]
        embedding = await client.batch_get_embedding(text)
        print(embedding)

    asyncio.run(main())


import aiohttp
import json
import time
from typing import List

class ModalEmbeddingClient():
    def __init__(self) -> None:
        self.web_url = "https://knight-h--bge-m3-bm25-model-get-embedding.modal.run"

    async def _get_embedding(self, text: str, is_query: bool = False) -> dict:
        """
        Sends a POST request to the embedding server to get the embedding for the given text.

        Args:
            text (str): The text to get the embedding for.
            is_query (bool): Whether the text is a query or not.

        Returns:
            dict: The embedding of the given text in the form of a dictionary.
        """
        start = time.time()
        async with aiohttp.ClientSession() as session:
            async with session.post(
                self.web_url,
                json={
                    "input": json.dumps({"doc_id":1, "document_text": text, "is_query": is_query}) + "\n",
                },
            ) as response:
                response_json = await response.json()

        print(json.dumps({
            "timestamp": time.strftime('[%Y-%b-%d %H:%M:%S %z]'),
            "message": "Modal serverless embedding latency",
            "latency": str(time.time() - start),
            "exception": False
        }))
        data = json.loads(response_json)
        # Get the embedding from the response
        data["embedding"] = data["dense_embeddings"] # keep forgetting to rename the keysss
        data["sparse_embedding"] = data["sparse_embeddings"]
        del data["dense_embeddings"], data["sparse_embeddings"]
        return data

    async def get_embedding(self, text: str, is_query: bool = False) -> dict:
        try:
            result = await self._get_embedding(text, is_query=is_query)
        except Exception as e:
            print(json.dumps({
                "timestamp": time.strftime('[%Y-%b-%d %H:%M:%S %z]'),
                "message": "ERROR Modal serverless embedding",
                "latency": None,
                "exception": True
            }))
            raise(e)
        return result

    async def batch_get_embedding(self, texts: List[str]) -> List[dict]:
        """
        Sends a POST request to the embedding server to get embeddings for a list of texts.
        Returns a list of embedding dicts.
        """
        start = time.time()
        lines = [
            json.dumps({"doc_id": i, "document_text": text}, ensure_ascii=False)
            for i, text in enumerate(texts)
        ]
        jsonl_body = "\n".join(lines) + "\n"
        async with aiohttp.ClientSession() as session:
            async with session.post(
                self.web_url,
                json={
                    "input": jsonl_body
                },
            ) as response:
                response_json = await response.json()

        print(json.dumps({
            "timestamp": time.strftime('[%Y-%b-%d %H:%M:%S %z]'),
            "message": "Modal serverless batch embedding latency",
            "latency": str(time.time() - start),
            "exception": False
        }))
        lines = response_json.strip().split('\n')
        data_list = []
        for line in lines:
            data = json.loads(line)
            data["embedding"] = data["dense_embeddings"]
            data["sparse_embedding"] = data["sparse_embeddings"]
            del data["dense_embeddings"], data["sparse_embeddings"]
            data_list.append(data)
        return data_list

if __name__=="__main__":
    import asyncio
    emb_client = ModalEmbeddingClient()
    text = "พนักงานทุกท่านซึ่งปฏิบัติงานมาครบ 1ปีจะได้รับวันหยุดประจำปี ดังนี้\n\nวันหยุดพักผ่อนประจำปี\n\nพนักงานทุกท่านจะได้รับวันหยุดพักผ่อนประจำปีท่านละ  10 วัน สำหรับพนักงานใหม่ที่ผ่านการทดลองงานแล้วก็ให้ลดลงตามส่วน\n\nถ้าลาเกินสิทธิ 10 วัน"

    res = asyncio.run(emb_client.get_embedding(text))
    assert "embedding" in res
    print(res)



from typing import Dict, Tuple, List
from collections import defaultdict
from fastapi import BackgroundTasks
import asyncio, time, json, uuid
from animuz_core.utils import init_logger

from openai import AsyncOpenAI

from animuz_core.genai.base import BaseAgent
from animuz_core.genai.prompts import CLAUDE_RAG_DEFAULT_PROMPT, CLAUDE_TOOL_DEFAULT_PROMPT
from animuz_core.genai.tools import OpenAITool

LOGGER = init_logger(__name__)

class OpenAIAgentClient(BaseAgent):
    def __init__(self, tools: Dict[str, Tuple[OpenAITool, callable]] = None) -> None:
        self.client = AsyncOpenAI()
        self.temperature = 0.0
        self.model = "gpt-4o"
        self.max_token = 1000
        if tools:
            self.add_tools(tools)

        self.result_store = {}

    def add_tools(self, tools: Dict[str, Tuple[OpenAITool, callable]]):
        self.tools = tools
        self.tool_dump = [{"type":"function", "function": tool[0].model_dump()} for tool in self.tools.values()]

    def _filter_for_openai(self, messages: List[Dict]) -> List[Dict]:
        """Remove non-OpenAI fields (metadata) before sending to API."""
        clean = []
        for msg in messages:
            m = msg.copy()
            m.pop("metadata", None)
            clean.append(m)
        return clean

    async def get_reply_frontend(self, messages: List[Dict[str, str]], user_chat_id: str) -> str:
        """
        Asynchronously retrieves a reply from the OpenAI Agent using the provided messages and system prompt.

        Args:
            messages (List[Dict[str, str]]): A list of dictionaries representing the messages in the conversation.
            system_prompt (str): The system prompt to use when creating the message.
            user_chat_id (str): The ID of the collection to use when using the retrieval tool.

        Returns:
            str: The content of the last message in the conversation.

        This function creates a message using the provided messages and system prompt, and appends the assistant's response to the messages list.
        If the message stop reason is not "tool_use", the assistant's response is appended to the output list.
        If the message stop reason is "tool_use", the function enters a loop where it processes the tool use response,
        appends the tool result to the messages and output lists, and creates a new message.
        The assistant's response for each new message is appended to the messages and output lists.
        The function returns the content of the last message in the conversation.
        """

        res = await self.create_message(messages)
        response_message = res.choices[0].message 
        tool_calls = response_message.tool_calls

        output = []

        messages.append(response_message)
        output.append(response_message)

        while tool_calls is not None:
            for tool_call in tool_calls:
                function_name = tool_call.function.name
                function_args = json.loads(tool_call.function.arguments)
                result, metadata = await self.process_tool_use(function_args, function_name, user_chat_id)

                tool_result = {
                    "tool_call_id": tool_call.id,
                    "role": "tool",
                    "name": function_name,
                    "content": result,
                    "metadata": metadata
                }
                messages.append(tool_result)
                output.append(tool_result)

            res = await self.create_message(messages)
            response_message = res.choices[0].message
            tool_calls = response_message.tool_calls

            messages.append(response_message)
            output.append(response_message)

        return output

    async def get_reply_frontend_tool_use_ack(self, messages: List[Dict[str, str]], user_chat_id: str, background_tasks: BackgroundTasks) -> str:

        res = await self.create_message(messages)
        response_message = res.choices[0].message 
        tool_calls = response_message.tool_calls

        output = []

        messages.append(response_message)
        output.append(response_message)

        if tool_calls is None:
            return output, None
        else:
            chat_id = str(uuid.uuid4()) 
            background_tasks.add_task(self.process_full_response, tool_calls, messages, user_chat_id, chat_id)
            return output, chat_id
        # message = await self.create_message(messages)
        # output = []

        # if message.stop_reason != "tool_use":
        #     messages.append({"role": "assistant", "content": message.content[-1].text})
        #     output.append({"role": "assistant", "content": message.content[-1].text})
        #     return output, None
        # else:
        #     messages.append({"role": "assistant", "content": message.content})
        #     output.append({"role": "assistant", "content": message.content})
        #     chat_id = str(uuid.uuid4()) 
        #     background_tasks.add_task(self.process_full_response, message, messages, user_chat_id, chat_id)

        #     return output, chat_id

    async def process_full_response(self, tool_calls, messages: List[Dict[str, str]], user_chat_id: str, chat_id: str) -> str:
        output = []

        while tool_calls is not None:
            for tool_call in tool_calls:
                function_name = tool_call.function.name
                function_args = json.loads(tool_call.function.arguments)
                result, metadata = await self.process_tool_use(function_args, function_name, user_chat_id)

                tool_result = {
                    "tool_call_id": tool_call.id,
                    "role": "tool",
                    "name": function_name,
                    "content": result,
                    "metadata": metadata
                }
                messages.append(tool_result)
                output.append(tool_result)

            res = await self.create_message(messages)
            response_message = res.choices[0].message
            tool_calls = response_message.tool_calls

            messages.append(response_message)
            output.append(response_message)

        self.result_store[chat_id] = output

    async def get_chat_result(self, chat_id: str):
            if chat_id in self.result_store:
                result = self.result_store[chat_id]
                del self.result_store[chat_id]  # Clean up the stored result
                return result
            else:
                return "Result not ready yet"
            
    async def create_message(self, history: List[Dict] = None, stream: bool= False):
        start = time.time()
        res = await self.client.chat.completions.create(
            model=self.model,
            max_tokens=self.max_token,
            temperature=self.temperature,
            messages=self._filter_for_openai(self.message_history if history is None else history),
            tools=self.tool_dump,
            stream=stream
        )

        LOGGER.debug(json.dumps({
            "timestamp": time.strftime('[%Y-%b-%d %H:%M:%S %z]'),
            "message": "OpenAI API latency",
            "latency": round(time.time() - start, 4),
            "exception": None
        }))
        return res

    async def stream_reply_frontend(self, messages: List[Dict[str, str]]):
        """Gets the response from OpenAI, updates the messages array, yields
        content, and calls functions as needed."""

        # Define variables to hold the streaming content and function call
        streaming_content = ""
        function_call = None
        content_started = False
        stream = await self.create_message(messages, stream = True)

        async for chunk in stream:
            msg = chunk.choices[0]
            # print(msg)
            if hasattr(msg, "delta"):
                if msg.delta.tool_calls:
                    if not function_call:
                        function_call = {"name": msg.delta.tool_calls[0].function.name, "arguments": "", "id": msg.delta.tool_calls[0].id}
                        yield {"type": "tool_call_start", "content": {"tool_name": msg.delta.tool_calls[0].function.name}}

                    if msg.delta.tool_calls[0].function.arguments:
                        function_call["arguments"] += msg.delta.tool_calls[0].function.arguments
                        yield {"type": "content", "content": msg.delta.tool_calls[0].function.arguments}                            

                # If it's content, add it to the streaming content and yield it
                elif hasattr(msg.delta, "content"):
                    if msg.delta.content:
                        streaming_content += msg.delta.content
                        if not content_started:
                            yield {"type":"content_start", "content": ""}
                            content_started = True
                        yield {"type":"content", "content":msg.delta.content}

            # If it's the end of the response and it's a text response, update the messages array with it
            if msg.finish_reason == "stop":
                yield {"type": "content_end", "content": streaming_content}

            elif msg.finish_reason == "tool_calls":
                yield {"type": "tool_call_end", "content": function_call}

        yield {"type": "message_end", "content": ""}

    async def get_reply_frontend_streaming(self, messages: List[Dict[str, str]], user_chat_id: str):
        output = []
        current_message = {"role": "assistant", "content": ""}
        tool_use = False
        async for chunk in self.stream_reply_frontend(messages):
            print(chunk)
            if chunk["type"] == "content":
                current_message["content"] += chunk["content"]
                yield chunk
            elif chunk["type"] == "content_start":
                yield chunk
            elif chunk["type"] == "content_end":
                yield chunk
            elif chunk["type"] == "tool_call_start":
                yield chunk
            elif chunk["type"] == "tool_call_end":
                yield chunk # yield tool call message

                # use tool
                tool_use = True
                result, metadata = await self.process_tool_use(json.loads(chunk["content"]["arguments"]), chunk["content"]["name"], user_chat_id)
                tool_result = {
                    "role": "function",
                    "tool_call_id": chunk["content"]["id"],
                    "content": result,
                    "name": chunk["content"]["name"],
                    "metadata": metadata
                }
                output.append(tool_result)
                messages.append({"role": "assistant", "content": None, "function_call":chunk["content"]}) # add tool call
                messages.append(tool_result) # add tool result

                yield {"type": "tool_result", "content": tool_result} # yield tool result message
            elif chunk["type"] == "message_end":
                if current_message["content"]:
                    output.append(current_message)
                yield {"type": "end", "content": output}

        while tool_use:
            current_message = {"role": "assistant", "content": ""}
            tool_use = False
            async for chunk in self.stream_reply_frontend(messages):
                if chunk["type"] == "content":
                    current_message["content"] += chunk["content"]
                    yield chunk
                elif chunk["type"] == "content_start":
                    yield chunk
                elif chunk["type"] == "content_end":
                    yield chunk
                elif chunk["type"] == "tool_call_start":
                    current_message["content"] += chunk["content"]
                    yield chunk
                elif chunk["type"] == "tool_call_end":
                    chunk["content"] = chunk["content"].to_dict()
                    yield chunk # yield tool call message

                    tool_use = True
                    result, metadata = await self.process_tool_use(json.loads(chunk["content"]["arguments"]), chunk["content"]["name"], user_chat_id)
                    tool_result = {
                        "role": "function",
                        "tool_call_id": chunk["content"]["id"],
                        "content": result,
                        "name": chunk["content"]["name"],
                        "metadata": metadata
                    }
                    output.append(tool_result)
                    messages.append({"role": "assistant", "content": None, "function_call":chunk["content"]}) # add tool call
                    messages.append(tool_result) # add tool result

                    yield {"type": "tool_result", "content": tool_result} # yield tool result message
                elif chunk["type"] == "message_end":
                    if current_message["content"]:
                        output.append(current_message)
                    yield {"type": "end", "content": output}

        if output:
            yield {"type": "final_output", "content": output}



import aiohttp
import asyncio
from typing import List, Callable, Optional

class HTTPEmbeddingClient():
    def __init__(
        self,
        host: str,
        port: Optional[int] = None,
        endpoint: str = "/",
        payload_builder: Optional[Callable[[str], dict]] = None,
        trust_env: bool = True,
    ) -> None:
        self.host = host
        self.port = port
        self.endpoint = endpoint.rstrip("/") or "/"
        self.payload_builder = payload_builder or (lambda text: {"text": text})
        self.trust_env = trust_env

        if self.port is not None:
            self.base_url = f"http://{self.host}:{self.port}"
        else:
            self.base_url = f"http://{self.host}"

    async def get_embedding(self, text: str) -> dict:
        """
        Sends a POST request to the embedding server to get the embedding for the given text.

        Args:
            text (str): The text to get the embedding for.

        Returns:
            dict: The embedding of the given text in the form of a dictionary.
        """
        async with aiohttp.ClientSession(trust_env=self.trust_env) as session:
            async with session.post(f"{self.base_url}{self.endpoint}", json=self.payload_builder(text)) as response:
                return await response.json()

    async def _abatch_get_embedding(self, texts) -> List[dict]:

        async def _a_get_embedding(session, text) -> dict:
            response = await session.post(self.endpoint, json=self.payload_builder(text))
            return await response.json(content_type=None)

        tasks = []
        async with aiohttp.ClientSession(self.base_url, trust_env=self.trust_env) as session:
            for text in texts:
                tasks.append(_a_get_embedding(session, text))

            result = await asyncio.gather(*tasks)
        return result

    async def batch_get_embedding(self, texts):
         return await self._abatch_get_embedding(texts)

# Backward-compatible alias
EmbeddingClient = HTTPEmbeddingClient


if __name__=="__main__":
    import asyncio, time

    async def main():
        emb_client = HTTPEmbeddingClient(host="localhost", port=12081)
        text = "พนักงานทุกท่านซึ่งปฏิบัติงานมาครบ 1ปีจะได้รับวันหยุดประจำปี ดังนี้\n\nวันหยุดพักผ่อนประจำปี\n\nพนักงานทุกท่านจะได้รับวันหยุดพักผ่อนประจำปีท่านละ  10 วัน สำหรับพนักงานใหม่ที่ผ่านการทดลองงานแล้วก็ให้ลดลงตามส่วน\n\nถ้าลาเกินสิทธิ 10 วัน"

        res = await emb_client.get_embedding(text)
        assert "embedding" in res

        start = time.time()
        for i in range(5):
            res = await emb_client.get_embedding(text)
        print("sequential time:", time.time()-start)
        print("--"*20)
        start = time.time()
        result = await emb_client.batch_get_embedding([text for i in range(5)])
        print(len(result))
        print("batch time:", time.time()-start)

    asyncio.run(main())


"""Animuz Core - Shared utilities for RAG system.

This package provides core utilities for building RAG (Retrieval-Augmented Generation) systems:
- LLM clients (OpenAI, Anthropic, Ollama)
- RAG pipelines (Simple and Agentic)
- Vector database integration (Qdrant)
- Embedding clients (Modal, S3, SageMaker)
- Document ingestion (Azure Doc AI, Unstructured, structured parsers)
"""

# Always-available imports (minimal deps: pydantic, httpx, aiohttp, requests)
from .genai.base import BaseLLM, BaseAgent
from .genai.tools import MCPTools
from .streaming.sse import format_sse_event, format_sse_error
from .streaming.message_mappers import map_to_responses_api, clean_output
from .tooling import ToolSpec, tool

try:
    from importlib.metadata import PackageNotFoundError, version
except ImportError:  # pragma: no cover - Python < 3.8
    PackageNotFoundError = Exception

    def version(_name):  # type: ignore[override]
        return "0.0.0"

try:
    __version__ = version("animuz-core")
except PackageNotFoundError:
    # Package metadata not available (e.g., editable without install)
    __version__ = "0.0.0"

# Lazy imports for modules with optional dependencies
_LAZY_IMPORTS = {
    # OpenAI (requires openai)
    "OpenAIAgentClient": ".genai.openai_client",
    "OpenAILLMClient": ".genai.openai_llm_client",
    "OpenAIAgentClientStreaming": ".genai.openai_streaming_client",
    "OpenAIAgentClientResponses": ".genai.openai_responses_client",
    # Anthropic (requires anthropic)
    "AnthropicClient": ".genai.anthropic_client",
    "AnthropicAgentClient": ".genai.anthropic_client",
    # Ollama (requires ollama)
    "OLlamaClient": ".genai.ollama_client",
    # Pipelines (requires aiobotocore)
    "BasePipeline": ".pipelines.base",
    "AgenticRAG": ".pipelines.agentic_rag",
    "SimpleRAG": ".pipelines.simple_rag",
    # Vector DB (requires qdrant-client)
    "QdrantDBClient": ".vectordb.utils",
    # Embedding
    "HTTPEmbeddingClient": ".embedding.embedding_client",
    "EmbeddingClient": ".embedding.embedding_client",
    "ModalEmbeddingClient": ".embedding.modal_embedding_client",
    "S3EmbeddingClient": ".embedding.s3_embedding_client",
    # Ingestion
    "AzureDocAiClient": ".ingest.azure_doc_ai_client",
    "MyUnstructuredClient": ".ingest.custom_unstructured_client",
    "Structured": ".ingest.structured",
    # Profiling
    "Profiler": ".profiling",
    # Unified RAG
    "RAG": ".rag",
    "MCP": ".rag",
    "build_prompt": ".rag",
    # Tooling factories
    "qdrant_retriever": ".tooling",
}

# Special name mappings (when the class name differs from the attr name in its module)
_LAZY_RENAMES = {
    "QdrantDBClient": "MyQdrantClient",
}


def __getattr__(name):
    if name in _LAZY_IMPORTS:
        import importlib
        module = importlib.import_module(_LAZY_IMPORTS[name], __name__)
        actual_name = _LAZY_RENAMES.get(name, name)
        return getattr(module, actual_name)
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


__all__ = [
    # Base classes
    "BaseLLM",
    "BaseAgent",
    "BasePipeline",
    # LLM clients
    "OpenAIAgentClient",
    "OpenAILLMClient",
    "OpenAIAgentClientStreaming",
    "OpenAIAgentClientResponses",
    "AnthropicClient",
    "AnthropicAgentClient",
    "OLlamaClient",
    # Tools
    "MCPTools",
    # Pipelines
    "AgenticRAG",
    "SimpleRAG",
    # Vector DB
    "QdrantDBClient",
    # Embedding
    "HTTPEmbeddingClient",
    "EmbeddingClient",
    "ModalEmbeddingClient",
    "S3EmbeddingClient",
    # Ingestion
    "AzureDocAiClient",
    "MyUnstructuredClient",
    "Structured",
    # Streaming helpers
    "format_sse_event",
    "format_sse_error",
    "map_to_responses_api",
    "clean_output",
    # Profiling
    "Profiler",
    # Unified RAG
    "RAG",
    "MCP",
    "build_prompt",
    "ToolSpec",
    "tool",
    "qdrant_retriever",
]



from abc import ABC, abstractmethod
import json, time
import logging

logger = logging.getLogger(__name__)

class BaseLLM(ABC):
    def clear_history(self) -> None:
        self.message_history = []

    @abstractmethod
    async def get_reply(self, prompt: str):
        pass

    def append_user_message(self, text: str):
        self.message_history.append({"role": "user", "content": text})

    def append_assistant_message(self, text: str):
        self.message_history.append({"role": "assistant", "content": text})

class BaseAgent(ABC):
    def clear_history(self) -> None:
        self.message_history = []

    @abstractmethod
    async def get_reply_frontend(self, prompt: str):
        pass

    def append_user_message(self, text: str):
        self.message_history.append({"role": "user", "content": text})

    def append_assistant_message(self, text: str):
        self.message_history.append({"role": "assistant", "content": text})

    async def process_tool_use(self, tool_input, tool_name, user_chat_id):
        try:
            if tool_name == "retriever":
                tool_input["user_chat_id"] = user_chat_id

            result = await self.tools[tool_name][1](**tool_input)

            if tool_name == "retriever":
                scores = [res.score for res in result[1]]
                payloads = [res.payload for res in result[1]]
                output = "\n".join([
                    f"<doc file=\"{_get_key(payload)}\">{result[0][i]}"
                    for i, payload in enumerate(payloads)
                ])
                metadata = json.dumps({
                    "documents": [{
                        "key": _get_key(payload),
                        "filetype": payload.get("filetype", ""),
                        "score": scores[i],
                    } for i, payload in enumerate(payloads)]
                }, ensure_ascii=False)
                return output, metadata
            return result, None
        except Exception as e:
            import traceback
            logger.warning("ERROR tool result for %s: %s", tool_name, e)
            traceback.print_exc()
            return "ERROR", None


def _get_key(payload: dict) -> str:
    """Extract display key from payload, removing uploads/{userID} prefix."""
    key = payload.get("key")
    if key:
        return "/".join(key.split("/")[2:])
    file_dir = payload.get("fileDirectory", "")
    return "/".join(file_dir.split("/")[2:])


import time, json
from typing import Union, List
from pathlib import Path

try:
    from azure.core.credentials import AzureKeyCredential
    from azure.ai.documentintelligence.aio import DocumentIntelligenceClient
    from azure.ai.documentintelligence.models import AnalyzeDocumentRequest, AnalyzeResult
    try:
        from azure.ai.documentintelligence.models import ContentFormat
    except ImportError:
        ContentFormat = None
except ImportError:
    AzureKeyCredential = None
    DocumentIntelligenceClient = None
    AnalyzeDocumentRequest = None
    ContentFormat = None
    AnalyzeResult = None

from langchain_text_splitters import MarkdownHeaderTextSplitter

from animuz_core.utils import init_logger

LOGGER = init_logger(__name__)

class AzureDocAiClient():
    def __init__(self, endpoint: str, key: str) -> None:
        self.client = DocumentIntelligenceClient(endpoint=endpoint, credential=AzureKeyCredential(key))

    async def analyze_document(self, file_dir: str, do_split: bool = True) -> Union[List[dict], str]:
        """
        Analyzes a document from the given file path.

        Args:
            file_dir (str): The path to the document file.
            do_split (bool; default=True): True if the document should be split with langchain markdown splitter.

        Returns:
            Union[List[dict], str]: If do_split is True: a list of dictionaries containing the text and metadata of each section of the document.
                                    If do_split is False: a string of the content of the analyzed document in markdown format.
        """
        start = time.time()
        with open(file_dir, "rb") as f:
            poller = await self.client.begin_analyze_document("prebuilt-layout", f, content_type="application/octet-stream", output_content_format=ContentFormat.MARKDOWN)
            
        result: AnalyzeResult = await poller.result()
        LOGGER.debug(json.dumps({
            "timestamp": time.strftime('[%Y-%b-%d %H:%M:%S %z]'),
            "message": "Azure Doc AI latency",
            "latency": round(time.time() - start, 4),
            "exception": False
        }))

        if not do_split:
            return result.content

        headers_to_split_on = [
            ("#", "Header 1"),
            ("##", "Header 2"),
        ]
        splitter = MarkdownHeaderTextSplitter(headers_to_split_on=headers_to_split_on,  strip_headers=False)
        splitted = splitter.split_text(result.content)
        res = []
        for s in splitted:
            res.append({"text": s.page_content, 
                        "source": "azure_doc_ai",
                        "file_directory": str(Path(file_dir).parent),
                        "metadata": {
                            "filename": file_dir,
                            }
                        })
            
        return res
    
if __name__=="__main__":
    import asyncio
    import dotenv, os
    dotenv.load_dotenv()

    client = AzureDocAiClient(endpoint=os.getenv("AZURE_DOC_AI_ENDPOINT"), key=os.getenv("AZURE_DOC_AI_KEY"))
    asyncio.run(client.analyze_document("test_docs/test.pdf"))


"""OpenAI Agent Client using the Responses API (non-streaming).

This is the recommended client for non-streaming use cases. It uses the newer
OpenAI Responses API (client.responses.create) instead of the legacy Chat
Completions API.
"""
from typing import Dict, List, Tuple
import time
import json
import logging

from openai import AsyncOpenAI

from animuz_core.genai.base import BaseAgent
from animuz_core.genai.tools import OpenAITool, MCPTools

logger = logging.getLogger(__name__)


def _truncate_for_log(history: List[Dict], max_content: int = 200) -> str:
    """Truncate long system messages and return JSON string for debug logging."""
    truncated = []
    for msg in (history or []):
        m = msg.copy()
        if m.get("role") == "system" and len(m.get("content", "")) > max_content:
            m["content"] = m["content"][:max_content] + "...[truncated]"
        truncated.append(m)
    return json.dumps(truncated, ensure_ascii=False)


class OpenAIAgentClientResponses(BaseAgent):
    """Non-streaming OpenAI Agent using the Responses API with tool loop.

    Supports both local tool dispatch and MCP-based tool execution.

    Args:
        user_chat_id: User/tenant ID for multi-tenant tool calls
        tools: Local tool dict (name -> (schema, callable))
        model: OpenAI model name (default: gpt-4o-mini)
        max_token: Max output tokens
        mcp_url: MCP server URL. If provided, tools are fetched from MCP.
        user_context: User context dict passed to MCP tool executions
        mcp_api_key: API key for MCP server. Defaults to MCP_API_KEY env var.
    """
    def __init__(
        self,
        user_chat_id: str = None,
        tools: Dict[str, Tuple[OpenAITool, callable]] = None,
        model: str = None,
        max_token: int = 2000,
        mcp_url: str = None,
        user_context: Dict = None,
        mcp_api_key: str = None,
        profiler=None,
        base_url: str = None,
    ) -> None:
        self.client = AsyncOpenAI(base_url=base_url)
        self.temperature = 0.0
        self.model = model if model else "gpt-4o-mini"
        self.max_token = max_token
        self.profiler = profiler
        self.user_chat_id = user_chat_id
        self.mcp = None
        if mcp_url:
            self.mcp = MCPTools(mcp_url, user_chat_id, user_context, mcp_api_key, profiler=profiler)
        elif tools:
            self.add_tools(tools)

    def add_tools(self, tools: Dict[str, Tuple[OpenAITool, callable]]):
        self.tools = tools
        self.tool_dump = [
            {"type": "function", **tool[0].model_dump()}
            for tool in self.tools.values()
        ]

    def _filter_for_openai(self, messages: List[Dict]) -> List[Dict]:
        """Remove non-OpenAI fields (metadata, complete) before sending to API."""
        clean_messages = []
        for msg in messages:
            clean_msg = msg.copy()
            clean_msg.pop("metadata", None)
            clean_msg.pop("complete", None)

            if clean_msg.get("type") == "message":
                clean_messages.append({
                    "type": "message",
                    "role": clean_msg.get("role"),
                    "content": clean_msg.get("content")
                })
            elif clean_msg.get("type") == "function_call":
                clean_messages.append({
                    "type": "function_call",
                    "name": clean_msg.get("name"),
                    "arguments": clean_msg.get("arguments"),
                    "call_id": clean_msg.get("call_id")
                })
            elif clean_msg.get("type") == "function_call_output":
                clean_messages.append({
                    "type": "function_call_output",
                    "call_id": clean_msg.get("call_id"),
                    "output": clean_msg.get("output")
                })
            else:
                clean_messages.append(clean_msg)

        return clean_messages

    async def get_reply_frontend(
        self,
        messages: List[Dict[str, str]],
        system_prompt: str = None
    ) -> List[Dict]:
        """Get a complete reply with automatic tool calling loop.

        Args:
            messages: Conversation messages (OpenAI Responses API format)
            system_prompt: System prompt to prepend

        Returns:
            List of output items (message dicts, function_call dicts,
            function_call_output dicts) representing the full conversation turn.
        """
        messages_with_prompt = [{"role": "system", "content": system_prompt}] + messages
        user_message = messages[-1]["content"]

        res = await self._create_message(self._filter_for_openai(messages_with_prompt))

        output = []
        has_function_call = False

        for item in res.output:
            if item.type == "message":
                msg_dict = {"type": "message", "role": "assistant", "content": res.output_text}
                messages_with_prompt.append(msg_dict)
                output.append(msg_dict)
            elif item.type == "function_call":
                has_function_call = True
                func_call_dict = {
                    "type": "function_call",
                    "name": item.name,
                    "arguments": item.arguments,
                    "call_id": item.call_id
                }
                messages_with_prompt.append(func_call_dict)
                output.append(func_call_dict)

        max_iterations = 10
        iteration_count = 0

        while has_function_call and iteration_count < max_iterations:
            iteration_count += 1

            for item in res.output:
                if item.type == "function_call":
                    args = json.loads(item.arguments)
                    function_name = item.name
                    logger.info(f"CALLING FUNCTION: {function_name} WITH ARGS: {str(args)}")

                    if self.mcp:
                        result, metadata = await self.process_tool_use_mcp(function_name, args, user_message)
                    else:
                        result, metadata = await self.process_tool_use(args, function_name, self.user_chat_id)

                    tool_result = {
                        "type": "function_call_output",
                        "call_id": item.call_id,
                        "output": result,
                        "metadata": metadata
                    }
                    output.append(tool_result)
                    messages_with_prompt.append(tool_result)

            res = await self._create_message(self._filter_for_openai(messages_with_prompt))

            has_function_call = False
            for item in res.output:
                if item.type == "message":
                    msg_dict = {"type": "message", "role": "assistant", "content": res.output_text}
                    messages_with_prompt.append(msg_dict)
                    output.append(msg_dict)
                elif item.type == "function_call":
                    has_function_call = True
                    func_call_dict = {
                        "type": "function_call",
                        "name": item.name,
                        "arguments": item.arguments,
                        "call_id": item.call_id
                    }
                    messages_with_prompt.append(func_call_dict)
                    output.append(func_call_dict)

        if iteration_count >= max_iterations:
            logger.warning(f"Max function call iterations ({max_iterations}) reached.")
            output.append({
                "type": "message",
                "role": "assistant",
                "content": "I apologize, but I've reached the maximum number of tool calls. Please try rephrasing your request."
            })

        return output

    async def _create_message(self, history: List[Dict]):
        """Call OpenAI Responses API."""
        start = time.time()
        logger.debug("Message History: " + _truncate_for_log(history))

        create_kwargs = dict(
            model=self.model,
            max_output_tokens=self.max_token,
            input=self.message_history if history is None else history,
            tools=self.mcp.tools if self.mcp else getattr(self, 'tool_dump', []),
        )
        if self.model not in ["gpt-5", "gpt-5-mini", "gpt-5-nano"]:
            create_kwargs["temperature"] = self.temperature

        res = await self.client.responses.create(**create_kwargs)

        openai_duration = time.time() - start
        logger.info(json.dumps({
            "timestamp": time.strftime('[%Y-%b-%d %H:%M:%S %z]'),
            "message": "OpenAI API latency",
            "latency": round(openai_duration, 4),
        }))
        if self.profiler:
            self.profiler.record_duration("openai_api", openai_duration)
        return res

    async def process_tool_use_mcp(self, tool_name: str, tool_input: dict, user_message: str):
        """Execute tool via MCP and return (result, metadata)."""
        start = time.time()
        try:
            response = self.mcp.execute_tool(tool_name, user_message=user_message, **tool_input)
            result, metadata = self.mcp.parse_response(tool_name, response, tool_input)

            logger.info(json.dumps({
                "timestamp": time.strftime('[%Y-%b-%d %H:%M:%S %z]'),
                "message": f"Tool {tool_name} used successfully",
                "latency": round(time.time() - start, 4),
            }, ensure_ascii=False))
            return result, metadata
        except Exception as e:
            logger.error(f"Tool {tool_name} error: {str(e)}")
            return "ERROR", None



import os
from pathlib import Path
from typing import List, Tuple

from animuz_core.pipelines.base import BasePipeline
from animuz_core.genai.base import BaseLLM
from animuz_core.embedding.embedding_client import EmbeddingClient
from animuz_core.ingest.custom_unstructured_client import MyUnstructuredClient
from animuz_core.ingest.azure_doc_ai_client import AzureDocAiClient
from animuz_core.vectordb.utils import MyQdrantClient

from animuz_core.utils import format_sources

class SimpleRAG(BasePipeline):
    def __init__(self, embedding_client: EmbeddingClient, db_client: MyQdrantClient, LLM: BaseLLM, unstructured_client: MyUnstructuredClient = None, azure_doc_ai_client: AzureDocAiClient = None) -> None:
        """
        Initializes the SimpleRAG object with the specified clients and models.

        Args:
            embedding_client (EmbeddingClient): The client for embedding data.
            db_client (MyQdrantClient): The client for the database.
            LLM (BaseLLM): The LLM model for the SimpleRAG object.
            unstructured_client (MyUnstructuredClient, optional): The client for unstructured data. If None, uses the default MyUnstructuredClient.
            azure_doc_ai_client (AzureDocAiClient, optional): The client for Azure Doc AI. If None, uses the default AzureDocAiClient.

        Returns:
            None: This function initializes the SimpleRAG object without returning anything.
        """
        super().__init__(embedding_client, db_client, unstructured_client, azure_doc_ai_client)

        self.llm = LLM

    async def add_doc(self, path: str, user_chat_id: str = "default") -> None:
        """
        Add a document to the database.

        Args:
            path (str): The path to the document.

        Returns:
            None: This function does not return anything.

        Raises:
            None: This function does not raise any exceptions.

        This function adds a document to the database by performing the following steps:
        1. Partition the document using the unstructured client or azure doc ai.
        2. Extract the properties of each partitioned data row.
        3. Get the embeddings for each partitioned data row using the embedding client.
        4. Extract the indices and values from the sparse embeddings.
        5. Upload the dense and sparse embeddings, indices, values, and properties to the database using the db client.

        Note: The document is expected to be in the specified path.
        """
        
        res = await self.parse_doc(path)

        properties = [{
                "text": data_row["text"],
                "source": data_row["source"],
                "file_directory": str(Path(path).parent),
                "filename": data_row["metadata"]["filename"],
                "filetype": data_row["metadata"]["filetype"],
                "userChatID": user_chat_id,
            } for data_row in res]
        
        emb_res = await self.batch_embed([str(data_row["text"]) for data_row in res])

        indices = [e["sparse_embedding"].keys() for e in emb_res]
        values = [e["sparse_embedding"].values() for e in emb_res]
            
        await self.upload_to_db([e["embedding"] for e in emb_res], indices, values, properties)

    async def query(self, query: str, user_chat_id: str) -> str:
        """
        Performs a query operation based on the input query string. Note that this is a stateless operation. No history is maintained.
        
        Args:
            query (str): The query string to search for.
            user_chat_id (str): The user chat ID used for multi-tenant retrieval.

        Returns:
            str: The response to the query.
        """
        search_results, _ = await self.query_db(query, user_chat_id)
        prompt = f"\n Use the following context to answer the user query: {format_sources(search_results)}. Be sure to include citations in your response (in the form of \"Source x: [answer]\" where x is an index). \n\n User query: {query}"
        reply = await self.llm.get_reply(prompt)
        self.llm.clear_history() #stateless chat
        return reply



"""Unified RAG API — 3 knobs: prompt, llm, tools.

Provides chat (with agent tool loop), document ingestion, and retrieval
through a single class.

Usage::

    from animuz_core import RAG, MCP, qdrant_retriever, tool

    # Local tools
    rag = RAG(
        prompt=http_fetcher,
        llm="gpt-4o-mini",
        tools=[qdrant_retriever(host="localhost", port=6333, collection="animuz")],
    )
    output = await rag.chat(team_id, assistant_id, messages)

    # MCP tools
    rag = RAG(prompt=ddb_fetcher, llm="gpt-4o-mini", tools=MCP(url=mcp_url))
    output = await rag.chat(team_id, assistant_id, messages, user_context=ctx)

    # Plain chat — no tools
    rag = RAG(prompt=my_fetcher, llm="gpt-4o-mini")
    output = await rag.chat(team_id, assistant_id, messages)

    # Ingest + retrieve (requires a qdrant_retriever tool)
    await rag.add_doc("docs/intro.md", user_chat_id="team1|asst1")
    texts, points = await rag.retrieve("what is this?", user_chat_id="team1|asst1")
"""

import inspect
import logging
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Callable, Dict, List, Optional, Union

from animuz_core.tooling import ToolSpec, RetrieverToolSpec, build_openai_tool

logger = logging.getLogger(__name__)


# ---------------------------------------------------------------------------
# MCP config
# ---------------------------------------------------------------------------

@dataclass
class MCP:
    """Configuration holder for an MCP tool server."""

    url: str
    api_key: Optional[str] = None


# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------

def build_prompt(assistant: dict) -> str:
    """Assemble a system prompt from an assistant config dict.

    Args:
        assistant: Must contain ``"prompt"`` key. Optional: ``"customPrompt"``.

    Returns:
        Assembled system prompt string including today's date.

    Raises:
        ValueError: If ``"prompt"`` key is missing.
    """
    if "prompt" not in assistant:
        raise ValueError("Assistant config missing 'prompt' key")

    prompt = assistant["prompt"]
    if "customPrompt" in assistant:
        prompt += "\n" + assistant["customPrompt"]
    prompt += f"\nToday is {datetime.now().strftime('%A, %d %B %Y')}."
    return prompt


def _resolve_provider(model: str) -> str:
    """Detect LLM provider from model name."""
    if model.startswith("claude"):
        return "anthropic"
    return "openai"


def _build_tool_dict(tools: List[ToolSpec]) -> Dict[str, tuple]:
    """Convert a list of ToolSpec into the ``{name: (OpenAITool, fn)}`` dict
    expected by ``OpenAIAgentClientResponses``."""
    return {spec.name: (build_openai_tool(spec), spec.fn) for spec in tools}


def _find_retriever(tools: Optional[Union[List[ToolSpec], "MCP"]]) -> Optional[RetrieverToolSpec]:
    """Return the first RetrieverToolSpec in the tools list, or None."""
    if not isinstance(tools, list):
        return None
    for t in tools:
        if isinstance(t, RetrieverToolSpec):
            return t
    return None


# ---------------------------------------------------------------------------
# RAG
# ---------------------------------------------------------------------------

class RAG:
    """Unified RAG interface with 3 knobs: **prompt**, **llm**, **tools**.

    Args:
        prompt: Callable ``(team_id, assistant_id) -> dict`` returning an
            assistant config dict. May be sync or async.
        llm: Default model name string. Provider is auto-detected from name.
        tools: ``list[ToolSpec]`` for local tools, ``MCP(...)`` for MCP
            server, or ``None`` for plain chat.
    """

    def __init__(
        self,
        prompt: Callable,
        *,
        llm: str = "gpt-4o-mini",
        tools: Optional[Union[List[ToolSpec], MCP]] = None,
        base_url: Optional[str] = None,
    ):
        self._fetcher = prompt
        self._default_model = llm
        self._tools = tools
        self._base_url = base_url

        # Pre-build local tool dict for the agent
        self._tool_dict: Optional[Dict[str, tuple]] = None
        if isinstance(tools, list):
            self._tool_dict = _build_tool_dict(tools)

        # Extract retriever clients for add_doc / retrieve
        self._retriever = _find_retriever(tools)

    # ------------------------------------------------------------------
    # Chat
    # ------------------------------------------------------------------

    async def _fetch_assistant(self, team_id: str, assistant_id: str) -> dict:
        """Call the user-provided fetcher, handling both sync and async."""
        result = self._fetcher(team_id, assistant_id)
        if inspect.isawaitable(result):
            result = await result
        return result

    async def chat(
        self,
        team_id: str,
        assistant_id: str,
        messages: List[Dict[str, str]],
        *,
        user_context: Optional[Dict] = None,
        profiler: Any = None,
    ) -> list:
        """Run the chat pipeline: fetch prompt → create agent → call agent → clean output.

        Args:
            team_id: Tenant / team identifier.
            assistant_id: Assistant identifier.
            messages: Conversation messages ``[{"role": ..., "content": ...}]``.
            user_context: Optional user context dict (passed to MCP tools).
            profiler: Optional profiler instance.

        Returns:
            Frontend-ready list of message dicts (via ``clean_output``).
        """
        from animuz_core.streaming.message_mappers import clean_output, map_to_responses_api

        # 1. Fetch assistant config
        assistant = await self._fetch_assistant(team_id, assistant_id)

        # 2. Build system prompt
        system_prompt = build_prompt(assistant)

        # 3. Resolve model (assistant override or default)
        model = assistant.get("model") or self._default_model

        # 4. Detect provider
        provider = _resolve_provider(model)
        if provider == "anthropic":
            raise NotImplementedError("Anthropic agent support is not yet implemented")

        # 5. Create agent
        from animuz_core.genai.openai_responses_client import OpenAIAgentClientResponses

        user_chat_id = f"{team_id}|{assistant_id}"

        agent_kwargs: Dict[str, Any] = {
            "user_chat_id": user_chat_id,
            "model": model,
            "profiler": profiler,
            "base_url": self._base_url,
        }

        if isinstance(self._tools, MCP):
            agent_kwargs["mcp_url"] = self._tools.url
            agent_kwargs["mcp_api_key"] = self._tools.api_key
            agent_kwargs["user_context"] = user_context
        elif self._tool_dict is not None:
            agent_kwargs["tools"] = self._tool_dict

        agent = OpenAIAgentClientResponses(**agent_kwargs)

        # 6. Map messages
        mapped = [map_to_responses_api(m) for m in messages]

        # 7. Get reply
        result = await agent.get_reply_frontend(mapped, system_prompt)

        # 8. Clean output
        return clean_output(result)

    # ------------------------------------------------------------------
    # Ingest
    # ------------------------------------------------------------------

    async def add_doc(self, path: str, user_chat_id: str) -> None:
        """Ingest a document: parse → embed → upload to vector DB.

        Requires a ``qdrant_retriever()`` tool in the tools list.

        Args:
            path: File path to ingest.
            user_chat_id: Tenant ID for multi-tenant isolation in the vector DB.

        Raises:
            RuntimeError: If no retriever tool is configured.
        """
        if not self._retriever:
            raise RuntimeError(
                "add_doc() requires a qdrant_retriever() tool. "
                "Pass tools=[qdrant_retriever(...)] when creating the RAG instance."
            )

        from animuz_core.pipelines.simple_rag import SimpleRAG
        from animuz_core.genai.base import BaseLLM

        # Create a minimal pipeline just for ingest (no LLM needed)
        class _NullLLM(BaseLLM):
            def __init__(self):
                self.clear_history()
            async def get_reply(self, prompt: str):
                raise RuntimeError("LLM not needed for ingest")

        pipeline = SimpleRAG(
            embedding_client=self._retriever.embedding_client,
            db_client=self._retriever.db_client,
            LLM=_NullLLM(),
        )
        await pipeline.add_doc(path, user_chat_id=user_chat_id)

    # ------------------------------------------------------------------
    # Retrieve
    # ------------------------------------------------------------------

    async def retrieve(self, query: str, user_chat_id: str, top_k: int = 3):
        """Retrieve documents from the vector DB (embed query → hybrid search).

        Requires a ``qdrant_retriever()`` tool in the tools list.

        Args:
            query: Search query string.
            user_chat_id: Tenant ID for multi-tenant filtering.
            top_k: Number of results to return.

        Returns:
            Tuple of (list of text strings, list of scored points).

        Raises:
            RuntimeError: If no retriever tool is configured.
        """
        if not self._retriever:
            raise RuntimeError(
                "retrieve() requires a qdrant_retriever() tool. "
                "Pass tools=[qdrant_retriever(...)] when creating the RAG instance."
            )

        return await self._retriever.fn(query, user_chat_id=user_chat_id, top_k=top_k)



import logging
import re
from typing import Literal, List, Tuple
from langchain_text_splitters import MarkdownTextSplitter, LatexTextSplitter, RecursiveCharacterTextSplitter

# ========== Interface ==========
class ResponseGetChunk:
    text: str
    source: str

# ========== Main Class ==========
class Structured():
    def __init__(self):
        pass

    @staticmethod
    def get_chunks(buffer: bytearray) -> Tuple[List[ResponseGetChunk], int]:
        text = clean_crlf(buffer.decode("utf-8"))
        char_count = len(text)

        rows = csv_to_array(text)   
        header = rows[0]
        content_index = header.index("content")

        if content_index == -1:
            logging.error('Content column not found')
            return [], -1
            
        chunks = [
            row[content_index]
            for row in rows[1:]
            if row[content_index]
        ]

        return [
            {
                "text": chunk,
                "source": "structured"
            }
            for chunk in chunks
        ], char_count

    @staticmethod
    def split_file(buffer: bytearray, file_extension: Literal["md", 'txt', 'tex', 'csv']) -> Tuple[List[ResponseGetChunk], int]:
        whole_file = clean_new_line(clean_crlf(buffer.decode("utf-8")))
        char_count = len(whole_file)

        if file_extension == 'md':
            return Structured.split_markdown(whole_file), char_count
        elif file_extension == 'txt' or file_extension == 'csv':
            return Structured.split_text(whole_file), char_count
        elif file_extension == 'tex':
            return Structured.split_latex(whole_file), char_count
        else:
            logging.error(f"Unsupported file extension: {file_extension}")
            return [], -1

    @staticmethod
    def split_markdown(text: str) -> List[ResponseGetChunk]:
        splitter = MarkdownTextSplitter(chunk_size=1200, chunk_overlap=200)
        split_result = splitter.split_text(text)
        output = [
            {
                "text": chunk,
                "source": "structured"
            }
            for chunk in split_result
        ]
        return output

    @staticmethod
    def split_text(text: str) -> List[ResponseGetChunk]:
        splitter = RecursiveCharacterTextSplitter(chunk_size=1200, chunk_overlap=200)
        split_result = splitter.split_text(text)
        output = [
            {
                "text": chunk,
                "source": "structured"
            }
            for chunk in split_result
        ]
        return output
    
    @staticmethod
    def split_latex(text: str) -> List[ResponseGetChunk]:
        splitter = LatexTextSplitter(chunk_size=1200, chunk_overlap=200)
        split_result = splitter.split_text(text)
        output = [
            {
                "text": chunk,
                "source": "structured"
            }
            for chunk in split_result
        ]
        return output

# ========== Function ==========
def clean_crlf(text: str) -> str:
    return re.sub(r'\r\n?|\n', '\n', text)

def clean_new_line(text: str) -> str:
    return re.sub(r'\n{3,}', '\n\n', text)

def csv_to_array(text: str) -> list[list[str]]:
    p = ""
    row = [""]
    ret = [row]
    i = 0
    r = 0
    s = True

    for l in text:
        if l == '"':
            if s and l == p:
                row[i] += l
            s = not s
        elif l == "," and s:
            l = ""
            row.append(l)
            i += 1
        elif l == "\n" and s:
            if p == "\r":
                row[i] = row[i][:-1]
            row = [""]
            ret.append(row)
            r += 1
            i = 0
        else:
            row[i] += l
        p = l


"""GenAI module - LLM clients and base classes."""

from .base import BaseLLM, BaseAgent
from .openai_client import OpenAIAgentClient
from .openai_streaming_client import OpenAIAgentClientStreaming
from .openai_responses_client import OpenAIAgentClientResponses
from .tools import (
    AnthropicTool,
    OpenAITool,
    AnthropicRetrieverTool,
    OpenAIRetrieverTool,
    MCPTools,
)

# Lazy imports for optional providers
def __getattr__(name):
    if name in ("AnthropicClient", "AnthropicAgentClient"):
        from .anthropic_client import AnthropicClient, AnthropicAgentClient
        return {"AnthropicClient": AnthropicClient, "AnthropicAgentClient": AnthropicAgentClient}[name]
    if name == "OLlamaClient":
        from .ollama_client import OLlamaClient
        return OLlamaClient
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")

__all__ = [
    "BaseLLM",
    "BaseAgent",
    "OpenAIAgentClient",
    "OpenAIAgentClientStreaming",
    "OpenAIAgentClientResponses",
    "AnthropicClient",
    "AnthropicAgentClient",
    "OLlamaClient",
    "AnthropicTool",
    "OpenAITool",
    "AnthropicRetrieverTool",
    "OpenAIRetrieverTool",
    "MCPTools",
]



"""Lightweight latency profiling for RAG pipelines.

Usage:
    from animuz_core.profiling import Profiler

    p = Profiler()
    p.mark("start")
    # ... do work ...
    p.mark("first_chunk")
    # ... more work ...
    p.mark("tool_call_start")
    p.mark("tool_call_end")
    p.mark("done")
    p.report()

Output:
    === Latency Profile ===
    start → first_chunk        1.234s
    first_chunk → tool_call_start  0.012s
    tool_call_start → tool_call_end  2.456s
    tool_call_end → done       0.089s
    ─────────────────────────
    TOTAL                      3.791s
"""
import time
import json
import logging
from typing import List, Tuple, Optional

logger = logging.getLogger(__name__)


class Profiler:
    """Simple mark-based latency profiler with counters and summaries."""

    def __init__(self, name: str = ""):
        self.name = name
        self.marks: List[Tuple[str, float]] = []
        self._start = time.perf_counter()
        self.counters: dict = {}  # e.g. {"tool_calls": 3}
        self.durations: dict = {}  # e.g. {"mcp_total": [0.5, 0.3]}

    def mark(self, label: str):
        """Record a timing mark."""
        self.marks.append((label, time.perf_counter()))

    def count(self, key: str, increment: int = 1):
        """Increment a named counter."""
        self.counters[key] = self.counters.get(key, 0) + increment

    def record_duration(self, key: str, duration: float):
        """Record a named duration for aggregation."""
        self.durations.setdefault(key, []).append(duration)

    def elapsed(self, label: str) -> Optional[float]:
        """Get elapsed time from start to a specific mark."""
        for name, t in self.marks:
            if name == label:
                return t - self._start
        return None

    def between(self, from_label: str, to_label: str) -> Optional[float]:
        """Get elapsed time between two marks."""
        t_from = t_to = None
        for name, t in self.marks:
            if name == from_label:
                t_from = t
            if name == to_label:
                t_to = t
        if t_from is not None and t_to is not None:
            return t_to - t_from
        return None

    def report(self) -> str:
        """Print and return a formatted latency report."""
        if len(self.marks) < 2:
            return "Not enough marks to report."

        lines = []
        header = f"=== Latency Profile{' (' + self.name + ')' if self.name else ''} ==="
        lines.append(header)

        max_label_len = max(
            len(f"{self.marks[i][0]} → {self.marks[i+1][0]}")
            for i in range(len(self.marks) - 1)
        )

        for i in range(len(self.marks) - 1):
            label = f"{self.marks[i][0]} → {self.marks[i+1][0]}"
            dt = self.marks[i+1][1] - self.marks[i][1]
            lines.append(f"  {label:<{max_label_len}}  {dt:.3f}s")

        total = self.marks[-1][1] - self.marks[0][1]
        lines.append("  " + "─" * (max_label_len + 8))
        lines.append(f"  {'TOTAL':<{max_label_len}}  {total:.3f}s")

        if self.counters:
            lines.append("")
            lines.append("  Counters:")
            for k, v in self.counters.items():
                lines.append(f"    {k}: {v}")

        if self.durations:
            lines.append("")
            lines.append("  Aggregated Durations:")
            for k, vals in self.durations.items():
                total_d = sum(vals)
                avg_d = total_d / len(vals)
                lines.append(f"    {k}: {total_d:.3f}s total, {avg_d:.3f}s avg ({len(vals)} calls)")

        report_str = "\n".join(lines)
        print(report_str)
        return report_str

    def to_dict(self) -> dict:
        """Export marks as a JSON-serializable dict."""
        result = {}
        for i in range(len(self.marks) - 1):
            key = f"{self.marks[i][0]}_to_{self.marks[i+1][0]}"
            result[key] = round(self.marks[i+1][1] - self.marks[i][1], 4)
        if len(self.marks) >= 2:
            result["total"] = round(self.marks[-1][1] - self.marks[0][1], 4)
        if self.counters:
            result["counters"] = dict(self.counters)
        if self.durations:
            result["durations"] = {
                k: {"total": round(sum(v), 4), "avg": round(sum(v)/len(v), 4), "count": len(v)}
                for k, v in self.durations.items()
            }
        return result

    def log(self):
        """Log the profile as structured JSON."""
        logger.info(json.dumps({
            "type": "latency_profile",
            "name": self.name,
            **self.to_dict()
        }))



"""
Copyright and License
----------------------
Copyright 2024 Artifex Software, Inc.
License GNU Affero GPL 3.0
"""
from pathlib import Path
from typing import List

import fitz

if fitz.pymupdf_version_tuple < (1, 24, 0):
    raise NotImplementedError("PyMuPDF version 1.24.0 or later is needed.")

def to_markdown(doc: fitz.Document, pages: list = None) -> str:
    """Process the document and return the text of its selected pages."""
    if isinstance(doc, str):
        doc = fitz.open(doc)
    if not pages:  # use all pages if argument not given
        pages = range(doc.page_count)

    md_string = ""

    for pno in pages:
        page = doc[pno]
        # 1. first locate all tables on page
        tabs = page.find_tables()

        # 2. make a list of table boundary boxes, sort by top-left corner.
        # Must include the header bbox, which may be external.
        tab_rects = sorted(
            [
                (fitz.Rect(t.bbox) | fitz.Rect(t.header.bbox), i)
                for i, t in enumerate(tabs.tables)
            ],
            key=lambda r: (r[0].y0, r[0].x0),
        )

        # 3. final list of all text and table rectangles
        text_rects = []
        # compute rectangles outside tables and fill final rect list
        for i, (r, idx) in enumerate(tab_rects):
            if i == 0:  # compute rect above all tables
                tr = page.rect
                tr.y1 = r.y0
                if not tr.is_empty:
                    text_rects.append(("text", tr, 0))
                text_rects.append(("table", r, idx))
                continue
            # read previous rectangle in final list: always a table!
            _, r0, idx0 = text_rects[-1]

            # check if a non-empty text rect is fitting in between tables
            tr = page.rect
            tr.y0 = r0.y1
            tr.y1 = r.y0
            if not tr.is_empty:  # empty if two tables overlap vertically!
                text_rects.append(("text", tr, 0))

            text_rects.append(("table", r, idx))

            # there may also be text below all tables
            if i == len(tab_rects) - 1:
                tr = page.rect
                tr.y0 = r.y1
                if not tr.is_empty:
                    text_rects.append(("text", tr, 0))

        if not text_rects:  # this will happen for table-free pages
            text_rects.append(("text", page.rect, 0))
        else:
            rtype, r, idx = text_rects[-1]
            if rtype == "table":
                tr = page.rect
                tr.y0 = r.y1
                if not tr.is_empty:
                    text_rects.append(("text", tr, 0))

        # we have all rectangles and can start outputting their contents
        for rtype, r, idx in text_rects:
            if rtype == "text":  # a text rectangle
                continue
            else:  # a table rect
                md_string += tabs[idx].to_markdown(clean=False)
                md_string += "</table>\n\n"

        md_string += "\n-----\n\n"

    return md_string


def extract_text_from_pdf(pdf_file: str) -> List[dict]:
    """
    Extracts text from a PDF file and returns a list of dictionaries containing the extracted text along with metadata.
    Tables are extracted as markdown but appended to the end of the text of each page instead of their original location. 
    Parameters:
        pdf_file (str): The path to the PDF file to extract text from.

    Returns:
        List[dict]: A list of dictionaries, each containing:
            - "text": The extracted text from the PDF page.
            - "source": The source of the text extraction (currently set to "pymupdf").
            - "file_directory": The directory of the PDF file.
            - "filename": The name of the PDF file without the extension.
            - "filetype": The extension of the PDF file.
            - "page_number": The page number corresponding to the extracted text.
    """

    all_res = []
    table_mds = to_markdown(pdf_file).split("\n-----\n\n") #this extracts only tables to markdown
    doc = fitz.open(pdf_file)
    page_no = 1
    for page, table_strs in zip(doc, table_mds):

        for tab in page.find_tables():
            # process the content of table 'tab'
            page.add_redact_annot(tab.bbox)  # wrap table in a redaction annotation

        page.apply_redactions()  # erase all table text
        page_text = page.get_text()

        if table_strs.strip():
            for table_str in table_strs.split("</table>\n\n"):
                page_text += "\n" + table_str + '\n'

        res = {
            "text": page_text.replace(" า", "ำ").replace(" ่า", "่ำ").replace(" ้า", "้ำ").replace(" ๊า", "๊ำ"),
            "source": "pymupdf",
            "file_directory": str(Path(pdf_file).parent),
            "filename": str(Path(pdf_file).stem),
            "filetype": Path(pdf_file).suffix,
            "page_number": page_no,
        }
        all_res.append(res)
        page_no+=1

    # TODO: add chunking
    return all_res


import asyncio, time, json
from concurrent.futures import ThreadPoolExecutor

from unstructured_client import UnstructuredClient
from unstructured_client.models import shared
from unstructured_client.models.errors import SDKError

from animuz_core.utils import init_logger

LOGGER = init_logger(__name__)

class MyUnstructuredClient():
    def __init__(self, host: str, port: int) -> None:
        self.client = UnstructuredClient(server_url=f"http://{host}:{port}", 
                                         api_key_auth=None,
                                        )
        
    def do_partition(self, req: shared.PartitionParameters):
        return self.client.general.partition(req)
    async def a_partition(self, filename: str):
        with open(filename, "rb") as f:
            # Note that this currently only supports a single file
            files=shared.Files(
                content=f.read(),
                file_name=filename,
            )

        req = shared.PartitionParameters(
            files=files,
            chunking_strategy="by_title",
            combine_under_n_chars=1500,
            max_characters=2000,
            new_after_n_chars=1500,
            # Other partition params
            strategy='auto',
            languages=["eng","tha"],
        )

        try:
            start = time.time()
            loop = asyncio.get_running_loop()
            with ThreadPoolExecutor() as pool:
                # Run the partition_document function asynchronously
                resp = await loop.run_in_executor(pool, self.do_partition, req)
                for e in resp.elements:
                    e["source"] = "unstructured"
                # LOGGER.debug(f"{time.strftime('[%Y-%b-%d %H:%M:%S %z]')} - Unstructured latency: {time.time() - start: .4f}s")
                LOGGER.debug(json.dumps({
                    "timestamp": time.strftime('[%Y-%b-%d %H:%M:%S %z]'),
                    "message": "Unstructured parsing latency",
                    "latency": str(round(time.time() - start, 4)),
                    "exception": False
                }))
                return resp.elements
        except SDKError as e:
            LOGGER.critical(json.dumps({
                "timestamp": time.strftime('[%Y-%b-%d %H:%M:%S %z]'),
                "message": "ERROR Unstructured parsing",
                "latency": None,
                "exception": True
            }), exc_info=True)
            # LOGGER.critical(f"{time.strftime('[%Y-%b-%d %H:%M:%S %z]')} - Unstructured error", exc_info=True)
            return None
        
class PaidUnstructuredClient(MyUnstructuredClient):
    def __init__(self, url: str, api_key: int) -> None:
        self.client = UnstructuredClient(server_url=url, 
                                         api_key_auth=api_key,
                                        )
        
if __name__ == "__main__":
    import time
    client = MyUnstructuredClient(host="192.168.1.60", port=8000)

    start = time.time()
    res = asyncio.run(client.a_partition("test_docs/test1.docx"))
    print(res)
    print(time.time() - start)


"""Message format conversion utilities.

Converts between different message formats (e.g., v2 chat messages to
OpenAI Responses API format).
"""


def map_to_responses_api(item: dict) -> dict:
    """Map a v2 chat message to OpenAI Responses API format.

    Handles three cases:
    - Messages with tool_calls -> function_call format
    - Tool role messages -> function_call_output format
    - Everything else -> pass through
    """
    if "tool_calls" in item:
        if not item["tool_calls"]:
            return item

        output = [{
            'type': 'function_call',
            'status': 'completed'
        }] * len(item["tool_calls"])

        for i, tool_call in enumerate(item["tool_calls"]):
            output[i]['arguments'] = tool_call['function']['arguments']
            output[i]['name'] = "rag" if tool_call['function']['name'] == "retriever" else tool_call['function']['name']
            output[i]['call_id'] = tool_call['id']
        return output[0]

    elif "role" in item:
        if item["role"] == "tool":
            output = {"type": "function_call_output"}
            output["call_id"] = item["tool_call_id"]
            output["output"] = str([item["content"], item["metadata"]])
            return output

    return item


def clean_output(serializable_output: list) -> list:
    """Transform raw agent output into frontend-friendly format.

    Groups function_call and function_call_output items as `debugPre` before
    each assistant message. Returns a list of dicts with keys:
    `role`, `content`, and `debugPre` (list of tool call/result dicts).

    Args:
        serializable_output: List of output items from an agent
            (message dicts, function_call dicts, function_call_output dicts)

    Returns:
        List of frontend-ready message dicts.
    """
    output = []
    debugPre = []
    for _message in serializable_output:
        if _message.get("type") == "message":
            content = _message.get("content", "")
            if isinstance(content, list):
                content = content[0]["text"]
            output.append({
                "debugPre": debugPre,
                "role": _message.get("role", "assistant"),
                "content": content,
            })
            debugPre = []
        else:
            debugPre.append(_message)
    return output



import time
import json

from openai import AsyncOpenAI

from animuz_core.genai.base import BaseLLM
from animuz_core.utils import init_logger


LOGGER = init_logger(__name__)


class OpenAILLMClient(BaseLLM):
    def __init__(self, system_prompt: str = None, model: str = "gpt-4o", max_token: int = 1000) -> None:
        self.client = AsyncOpenAI()
        self.system_prompt = system_prompt
        self.model = model
        self.max_token = max_token
        self.temperature = 0.0
        self.message_history = []

    async def get_reply(self, prompt: str):
        messages = []
        if self.system_prompt:
            messages.append({"role": "system", "content": self.system_prompt})
        messages.append({"role": "user", "content": prompt})

        start = time.time()
        res = await self.client.chat.completions.create(
            model=self.model,
            max_tokens=self.max_token,
            temperature=self.temperature,
            messages=messages,
        )
        LOGGER.debug(json.dumps({
            "timestamp": time.strftime('[%Y-%b-%d %H:%M:%S %z]'),
            "message": "OpenAI LLM latency",
            "latency": round(time.time() - start, 4),
            "exception": None,
        }))
        return res.choices[0].message.content



"""Server-Sent Events (SSE) formatting helpers."""
import json


def format_sse_event(data: dict) -> str:
    """Format a dict as an SSE data line."""
    return f"data: {json.dumps(data, ensure_ascii=False)}\n\n"


def format_sse_error(message: str, status_code: int) -> str:
    """Format an error as an SSE event."""
    error_data = {
        "type": "error",
        "message": message,
        "statusCode": status_code
    }
    return f"data: {json.dumps(error_data, ensure_ascii=False)}\n\n"



import inspect
from dataclasses import dataclass
from typing import Any, Callable, Dict, Optional

from animuz_core.genai.tools import InputSchema, Property, OpenAITool, AnthropicTool


@dataclass
class ToolSpec:
    name: str
    description: str
    parameters: Dict[str, Any]
    fn: Callable[..., Any]


@dataclass
class RetrieverToolSpec(ToolSpec):
    """ToolSpec that also carries its embedding and DB clients.

    Returned by ``qdrant_retriever()``. ``RAG`` detects this subclass to
    enable ``add_doc()`` and ``retrieve()`` without extra configuration.
    """

    embedding_client: Any = None
    db_client: Any = None


def _infer_parameters(fn: Callable[..., Any]) -> Dict[str, Any]:
    signature = inspect.signature(fn)
    properties: Dict[str, Dict[str, str]] = {}
    required = []

    type_map = {
        str: "string",
        int: "integer",
        float: "number",
        bool: "boolean",
    }

    for param in signature.parameters.values():
        if param.kind in (inspect.Parameter.VAR_KEYWORD, inspect.Parameter.VAR_POSITIONAL):
            continue

        annotation = param.annotation
        json_type = type_map.get(annotation, "string")
        properties[param.name] = {
            "type": json_type,
            "description": "",
        }

        if param.default is inspect.Parameter.empty:
            required.append(param.name)

    return {
        "type": "object",
        "properties": properties,
        "required": required,
    }


def tool(
    fn: Optional[Callable[..., Any]] = None,
    *,
    name: Optional[str] = None,
    description: Optional[str] = None,
    parameters: Optional[Dict[str, Any]] = None,
):
    def wrap(func: Callable[..., Any]) -> ToolSpec:
        return ToolSpec(
            name=name or func.__name__,
            description=description or (func.__doc__ or ""),
            parameters=parameters or _infer_parameters(func),
            fn=func,
        )

    return wrap if fn is None else wrap(fn)


def _schema_to_input_schema(schema: Dict[str, Any]) -> InputSchema:
    properties = {
        key: Property(type=value.get("type", "string"), description=value.get("description", ""))
        for key, value in schema.get("properties", {}).items()
    }
    return InputSchema(
        type=schema.get("type", "object"),
        properties=properties,
        required=schema.get("required", []),
    )


def build_openai_tool(spec: ToolSpec) -> OpenAITool:
    return OpenAITool(
        name=spec.name,
        description=spec.description,
        parameters=_schema_to_input_schema(spec.parameters),
    )


def build_anthropic_tool(spec: ToolSpec) -> AnthropicTool:
    return AnthropicTool(
        name=spec.name,
        description=spec.description,
        input_schema=_schema_to_input_schema(spec.parameters),
    )


def qdrant_retriever(
    host: str = "localhost",
    port: int = 6333,
    collection: str = "animuz",
    *,
    embedding_client=None,
    api_key: Optional[str] = None,
) -> ToolSpec:
    """Factory that returns a ToolSpec named ``retriever`` backed by Qdrant hybrid search.

    The returned tool declares only a ``query`` parameter (what the LLM sees).
    ``user_chat_id`` is injected at runtime by ``BaseAgent.process_tool_use``.

    Args:
        host: Qdrant host.
        port: Qdrant port.
        collection: Qdrant collection name.
        embedding_client: Override the default ``ModalEmbeddingClient``.
        api_key: Qdrant API key (also reads ``QDRANT_API_KEY`` env var).
    """
    import os as _os
    from animuz_core.embedding.modal_embedding_client import ModalEmbeddingClient
    from animuz_core.vectordb.utils import MyQdrantClient

    _emb = embedding_client or ModalEmbeddingClient()
    _db = MyQdrantClient(host=host, port=port, collection_name=collection)

    async def query_db(query: str, user_chat_id: str, hybrid: bool = True, top_k: int = 3):
        emb_res = await _emb.get_embedding(query)
        indices = emb_res["sparse_embedding"].keys()
        values = emb_res["sparse_embedding"].values()
        search_results = await _db.hybrid_search(
            emb_res["embedding"], indices, values,
            top_k=top_k, user_chat_id=user_chat_id,
        )
        return [res.payload["text"] for res in search_results.points], search_results.points

    return RetrieverToolSpec(
        name="retriever",
        description="Retrieve documents related to a given query from a vector database",
        parameters={
            "type": "object",
            "properties": {
                "query": {
                    "type": "string",
                    "description": "The user query to be used for retrieval. The query must be comprehensive and in Thai",
                },
            },
            "required": ["query"],
        },
        fn=query_db,
        embedding_client=_emb,
        db_client=_db,
    )



"""Ingest module - Document ingestion and parsing utilities."""

from .azure_doc_ai_client import AzureDocAiClient
from .custom_unstructured_client import MyUnstructuredClient, PaidUnstructuredClient
from .structured import Structured

__all__ = [
    "AzureDocAiClient",
    "MyUnstructuredClient",
    "PaidUnstructuredClient",
    "Structured",
]
